import json
import logging
import re
import warnings
from typing import Dict, List, Literal, Union

from parsons import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)

API_URL = "https://actionnetwork.org/api/v2"


class ActionNetwork(object):
    """
    `Args:`
        api_token: str
            The OSDI API token
    """

    def __init__(self, api_token=None):
        self.api_token = check_env.check("AN_API_TOKEN", api_token)
        self.headers = {
            "Content-Type": "application/json",
            "OSDI-API-Token": self.api_token,
        }
        self.api_url = API_URL
        self.api = APIConnector(self.api_url, headers=self.headers)

    def _get_page(self, object_name, page, per_page=25, filter=None):
        # returns data from one page of results
        if per_page > 25:
            per_page = 25
            logger.info(
                "Action Network's API will not return more than 25 entries per page. \
            Changing per_page parameter to 25."
            )
        params = {"page": page, "per_page": per_page, "filter": filter}
        return self.api.get_request(url=object_name, params=params)

    def _get_entry_list(self, object_name, limit=None, per_page=25, filter=None):
        # returns a list of entries for a given object, such as people, tags, or actions
        # Filter can only be applied to people, petitions, events, forms, fundraising_pages,
        # event_campaigns, campaigns, advocacy_campaigns, signatures, attendances, submissions,
        # donations and outreaches.
        # See Action Network API docs for more info: https://actionnetwork.org/docs/v2/
        count = 0
        page = 1
        return_list = []
        while True:
            response = self._get_page(object_name, page, per_page, filter=filter)
            page = page + 1
            response_list = response["_embedded"][list(response["_embedded"])[0]]
            if not response_list:
                return Table(return_list)
            return_list.extend(response_list)
            count = count + len(response_list)
            if limit:
                if count >= limit:
                    return Table(return_list[0:limit])

    # Advocacy Campaigns
    def get_advocacy_campaigns(self, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.

        `Returns:`
            A  JSON with all of the advocacy_campaigns (letters) entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/advocacy_campaigns
        """
        if page:
            return self._get_page("advocacy_campaigns", page, per_page, filter)
        return self._get_entry_list("advocacy_campaigns", limit, per_page, filter)

    def get_advocacy_campaign(self, advocacy_campaign_id):
        """
        `Args:`
            advocacy_campaign_id:
               The unique id of the advocacy_campaign
        `Returns:`
            A  JSON with advocacy_campaign entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/advocacy_campaigns
        """
        return self.api.get_request(f"advocacy_campaigns/{advocacy_campaign_id}")

    # Attendances
    def get_person_attendances(self, person_id, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            person_id:
               The unique id of the person
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.
        `Returns:`
            A  JSON with all the attendances entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/attendances
        """
        if page:
            return self._get_page(f"people/{person_id}/attendances", page, per_page, filter)
        return self._get_entry_list(f"people/{person_id}/attendances", limit, per_page, filter)

    def get_event_attendances(self, event_id, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            event_id: the unique id of the event
            limit:
               The number of entries to return. When None, returns all entries.
           per_page
               The number of entries per page to return. 25 maximum.
           page
               Which page of results to return
           filter
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.
        `Returns:`
            A  JSON with the attendances entries related to the event
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/attendances
        """
        if page:
            return self._get_page(f"events/{event_id}/attendances", page, per_page, filter)
        return self._get_entry_list(f"events/{event_id}/attendances", limit, per_page, filter)

    def get_event_attendance(self, event_id, attendance_id):
        """
        `Args:`
            event_id:
               The unique id of the event
            attendance_id:
               The unique id of the attendance
        `Returns:`
            A  JSON with the attendance entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/attendances
        """
        return self.api.get_request(f"events/{event_id}/attendances/{attendance_id}")

    def get_person_attendance(self, person_id, attendance_id):
        """
        `Args:`
            person_id:
               The unique id of the person
            attendance_id:
               The unique id of the attendance
        `Returns:`
            A  JSON with the attendance entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/attendances
        """
        return self.api.get_request(f"people/{person_id}/attendances/{attendance_id}")

    def create_attendance(self, event_id, payload):
        """
        `Args:`
            event_id: the unique id of the event
            payload: the payload for creating the event attendance
                {
                    "_links" : {
                        "osdi:person" : { "href" : "https://actionnetwork.org/api/v2/people/id" }
                    }
                }
        `Returns:`
            A JSON response after creating the event attendance
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/attendances
        """
        return self.api.post_request(f"events/{event_id}/attendances", payload)

    def update_attendance(self, event_id, attendance_id, payload):
        """
        `Args:`
            event_id:
               The unique id of the event
            attendance_id:
               The unique id of the attendance
            payload:
               The payload for updating the event attendance
               {
                   "identifiers": [
                       "other-system:230125a"
                   ]
               }
        `Returns:`
            A JSON response after updating the event attendance
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/attendances
        """
        return self.api.put_request(f"events/{event_id}/attendances/{attendance_id}", payload)

    # Campaigns
    def get_campaigns(self, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.
        `Returns:`
            A  JSON with all of the campaigns entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/campaigns
        """
        if page:
            return self._get_page("campaigns", page, per_page, filter)
        return self._get_entry_list("campaigns", limit, per_page, filter)

    def get_campaign(self, campaign_id):
        """
        `Args:`
            campaign_id:
               The unique id of the campaign
        `Returns:`
            A  JSON with the campaign entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/campaigns
        """
        return self.api.get_request(f"campaigns/{campaign_id}")

    # Custom Fields
    def get_custom_fields(self):
        """
        `Args:`
            None
        `Returns:`
            A  JSON with the custom_fields associated with your API key.
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/custom_fields
        """
        return self.api.get_request("metadata/custom_fields")

    # Donations
    def get_donation(self, donation_id):
        """
        `Args:`
            donation_id: The unique id of the donation
        `Returns:`
            A  JSON with donation data
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/donations
        """
        return self.api.get_request(url=f"donations/{donation_id}")

    def get_donations(self, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
             limit:
                The number of entries to return. When None, returns all entries.
            per_page:
                The number of entries per page to return. 25 maximum.
            page:
                Which page of results to return
            filter:
                The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
                When None, no filter is applied.
        `Returns:`
            A  JSON with all the donations entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/donations
        """
        if page:
            return self._get_page("donations", page, per_page, filter)
        return self._get_entry_list("donations", limit, per_page, filter)

    def get_fundraising_page_donations(
        self, fundraising_page_id, limit=None, per_page=25, page=None, filter=None
    ):
        """
        `Args:`
             fundraising_page_id: The id of the fundraiser
             limit:
                The number of entries to return. When None, returns all entries.
            per_page
                The number of entries per page to return. 25 maximum.
            page
                Which page of results to return
            filter
                The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
                When None, no filter is applied.

        `Returns:`
            A  JSON with fundraising_page entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/donations
        """
        if page:
            return self._get_page(
                f"fundraising_pages/{fundraising_page_id}/donations",
                page,
                per_page,
                filter,
            )
        return self._get_entry_list(
            f"fundraising_pages/{fundraising_page_id}/donations",
            limit,
            per_page,
            filter,
        )

    def get_person_donations(self, person_id, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
             person_id: The id of the person
             limit:
                The number of entries to return. When None, returns all entries.
            per_page
                The number of entries per page to return. 25 maximum.
            page
                Which page of results to return
            filter
                The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
                When None, no filter is applied.

        `Returns:`
            A  JSON with all donations related to person
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/donations
        """
        if page:
            return self._get_page(
                f"people/{person_id}/donations",
                page,
                per_page,
                filter,
            )
        return self._get_entry_list(
            f"people/{person_id}/donations",
            limit,
            per_page,
            filter,
        )

    def create_donation(self, fundraising_page_id, donation_payload):
        """
        `Args:`
            fundraising_page_id: The id of the fundraising page
            donation_payload: The payload containing donation details
                {
                    "recipients": [
                        {
                            "display_name": "Campaign To Elect Tom",
                            "amount": "3.00"
                        }
                    ],
                    "created_date": "2013-01-01T00:00:00Z",
                    "_links" : {
                        "osdi:person" : { "href" : "link" }
                    }
                }
        `Returns:`
            A JSON response confirming the creation of the donation
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/donations
        """
        return self.api.post_request(
            f"fundraising_pages/{fundraising_page_id}/donations", donation_payload
        )

    # Embeds
    def get_embeds(self, action_type, action_id):
        """
        `Args:`
            action_type:
              The action type (petition, events, etc.)
            action_id:
              The unique id of the action
        `Returns:`
            A  JSON with the embeds (for you to be able to embed action outside of ActionNetwork).
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/embeds
        """
        return self.api.get_request(f"{action_type}/{action_id}/embed")

    # Event Campaigns
    def get_event_campaigns(self, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.
        `Returns:`
            A  JSON with all the event_campaigns entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/event_campaigns
        """
        if page:
            return self._get_page("event_campaigns", page, per_page, filter)
        return self._get_entry_list("event_campaigns", limit, per_page, filter)

    def get_event_campaign(self, event_campaign_id):
        """
        `Args:`
            event_campaign_id:
               The unique id of the event_campaign
        `Returns:`
            A  JSON with event_campaign entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/event_campaigns
        """
        return self.api.get_request(f"event_campaigns/{event_campaign_id}")

    def create_event_campaign(self, payload):
        """
        `Args:`
            payload: The payload containing event campaign details
                {
                    "title": "My Canvassing Event",
                    "origin_system": "CanvassingEvents.com"
                }
        `Returns:`
            A JSON response confirming the creation of the event campaign
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/event_campaigns
        """
        return self.api.post_request("event_campaigns", payload)

    def create_event_in_event_campaign(self, event_campaign_id, payload):
        """
        `Args:`
            event_campaign_id:
               The unique id of the event_campaign
            payload:
               The payload containing event details
               {
                   "title": "My Free Event",
                   "origin_system": "FreeEvents.com"
               }
        `Returns:`
            A JSON response confirming the creation of the event in the event campaign
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/event_campaigns
        """
        return self.api.post_request(f"event_campaigns/{event_campaign_id}/events", payload)

    def update_event_campaign(self, event_campaign_id, payload):
        """
        `Args:`
            event_campaign_id:
               The unique id of the event_campaign
            payload:
               The payload containing event campaign details
               {
                   "description": "This is my new event campaign description"
               }
        `Returns:`
            A JSON response confirming the update of the event campaign
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/event_campaigns
        """
        return self.api.put_request(f"event_campaigns/{event_campaign_id}", payload)

    # Events
    def get_events(self, limit=None, per_page=25, page=None, filter=None):
        """
         `Args:`
        limit:
            The number of entries to return. When None, returns all entries.
        per_page
            The number of entries per page to return. 25 maximum.
        page
            Which page of results to return
        filter
            The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
            When None, no filter is applied.
         `Returns:`
             A  JSON with all the events entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/events
        """
        if page:
            return self._get_page("events", page, per_page, filter)
        return self._get_entry_list("events", limit, per_page, filter)

    def get_event(self, event_id):
        """
        `Args:`
            event_id: the unique id of the event
        `Returns:`
            A  JSON with event entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/events
        """
        return self.api.get_request(f"events/{event_id}")

    def get_event_campaign_events(
        self, event_campaign_id, limit=None, per_page=25, page=None, filter=None
    ):
        """
        `Args:`
            event_campaign_id:
               The unique id of the event_campaign
            limit:
                The number of entries to return. When None, returns all entries.
            per_page
                The number of entries per page to return. 25 maximum.
            page
                Which page of results to return
            filter
                The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
                When None, no filter is applied.
        `Returns:`
            A  JSON with all the eventes related to the event_campaign entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/events
        """
        if page:
            return self._get_page(
                f"event_campaigns/{event_campaign_id}/events", page, per_page, filter
            )
        return self._get_entry_list(
            f"event_campaigns/{event_campaign_id}/events", limit, per_page, filter
        )

    def create_event(self, title, start_date=None, location=None):
        """
        Create an event in Action Network

        `Args:`
            title: str
                The public title of the event
            start_date: str OR datetime
                OPTIONAL: The starting date & time. If a string, use format "YYYY-MM-DD HH:MM:SS"
                (hint: the default format you get when you use `str()` on a datetime)
            location: dict
                OPTIONAL: A dict of location details. Can include any combination of the types of
                values in the following example:
                .. code-block:: python

                    my_location = {
                        "venue": "White House",
                        "address_lines": [
                            "1600 Pennsylvania Ave"
                        ],
                        "locality": "Washington",
                        "region": "DC",
                        "postal_code": "20009",
                        "country": "US"
                    }

        `Returns:`
            Dict of Action Network Event data.
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/events
        """

        data = {"title": title}

        if start_date:
            start_date = str(start_date)
            data["start_date"] = start_date

        if isinstance(location, dict):
            data["location"] = location

        event_dict = self.api.post_request(url=f"{self.api_url}/events", data=json.dumps(data))

        an_event_id = event_dict["_links"]["self"]["href"].split("/")[-1]
        event_dict["event_id"] = an_event_id

        return event_dict

    def update_event(self, event_id, payload):
        """
        Update an event in Action Network

        `Args:`
            event_id: str
                The unique id of the event
            payload: dict
                The payload containing event data (see https://actionnetwork.org/docs/v2/events)
                {
                    "title": "My Free Event With A New Name",
                    "description": "This is my free event description"
                }

        `Returns:`
            A JSON response confirming the update of the event
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/events
        """
        return self.api.put_request(f"events/{event_id}", payload)

    # Forms
    def get_forms(self, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
             limit:
                The number of entries to return. When None, returns all entries.
            per_page
                The number of entries per page to return. 25 maximum.
            page
                Which page of results to return
            filter
                The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
                When None, no filter is applied.
        `Returns:`
            A  JSON with all the forms entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/forms
        """
        if page:
            return self._get_page("forms", page, per_page, filter)
        return self._get_entry_list("forms", limit, per_page, filter)

    def get_form(self, form_id):
        """
        `Args:`
            form_id:
               The unique id of the form
        `Returns:`
            A  JSON with form entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/forms
        """
        return self.api.get_request(f"forms/{form_id}")

    def create_form(self, payload):
        """
        Create a form in Action Network

        `Args:`
            payload: dict
                The payload containing form details

                {
                    "title": "My Free Form",
                    "origin_system": "FreeForms.com"
                }

        `Returns:`
            A JSON response confirming the creation of the form
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/forms
        """
        return self.api.post_request("forms", payload)

    def update_form(self, form_id, payload):
        """
        Update a form in Action Network

        `Args:`
            form_id:
                The unique id of the form
            payload: dict
                The payload containing form data (see https://actionnetwork.org/docs/v2/forms)

                {
                    "title": "My Free Form",
                    "origin_system": "FreeForms.com"
                }

        `Returns:`
            A JSON response confirming the update of the form
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/forms
        """
        return self.api.put_request(f"forms/{form_id}", payload)

    # Fundraising Pages
    def get_fundraising_page(self, fundraising_page_id):
        """
        `Args:`
            fundraising_page_id: The id of the fundraiser
        `Returns:`
            A  JSON with fundraising_page entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/fundraising_pages
        """
        return self.api.get_request(url=f"fundraising_pages/{fundraising_page_id}")

    def get_fundraising_pages(self, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
             limit:
                The number of entries to return. When None, returns all entries.
            per_page
                The number of entries per page to return. 25 maximum.
            page
                Which page of results to return
            filter
                The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
                When None, no filter is applied.

        `Returns:`
            A  JSON with all the fundraising_pages entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/fundraising_pages
        """
        if page:
            return self._get_page("fundraising_pages", page, per_page, filter)
        return self._get_entry_list(
            "fundraising_pages",
            limit,
        )

    def create_fundraising_page(self, payload):
        """
        Create a fundraising page in Action Network

        `Args:`
            payload: dict
                The payload containing fundraising page details
                {
                    "title": "My Free Fundraiser",
                    "origin_system": "FreeFundraisers.com"
                }
        `Returns:`
            A JSON response confirming the creation of the fundraising page
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/fundraising_pages
        """
        return self.api.post_request("fundraising_pages", payload)

    def update_fundraising_page(self, fundraising_page_id, payload):
        """
        Update a fundraising page in Action Network

        `Args:`
            fundraising_page_id: The id of the fundraiser
            payload: dict
                The payload containing updated fundraising page details
                {
                    "title": "My Free Fundraiser",
                    "origin_system": "FreeFundraisers.com"
                }
        `Returns:`
            A JSON response confirming the update of the fundraising page
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/fundraising_pages
        """
        return self.api.put_request(f"fundraising_pages/{fundraising_page_id}", payload)

    # Items
    def get_items(self, list_id, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            list_id:
                The unique id of the list
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.
        `Returns:`
            A  JSON with all the list item entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/items
        """
        if page:
            return self._get_page(f"lists/{list_id}/items", page, per_page, filter)
        return self._get_entry_list(f"lists/{list_id}/items", limit, per_page, filter)

    def get_item(self, list_id, item_id):
        """
        `Args:`
           list_id:
              The unique id of the list
            item_id:
              The unique id of the item
        `Returns:`
            A  JSON with the item entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/items
        """
        return self.api.get_request(f"lists/{list_id}/items/{item_id}")

    # Lists
    def get_lists(self, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.


        `Returns:`
            A  JSON with all the list entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/lists
        """
        if page:
            return self._get_page("lists", page, per_page, filter)
        return self._get_entry_list("lists", limit, per_page, filter)

    def get_list(self, list_id):
        """
        `Args:`
           list_id:
              The unique id of the list
        `Returns:`
            A  JSON with the list entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/lists
        """
        return self.api.get_request(f"lists/{list_id}")

    # Messages
    def get_messages(
        self, limit=None, per_page=25, page=None, filter=None, unpack_statistics=False
    ):
        """
        `Args:`
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.
           unpack_statistics:
               Whether to unpack the statistics dictionary into the table. Default to False.

        `Returns:`
            A Parsons Table with all the messages related entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/messages
        """
        if page:
            return self._get_page("messages", page, per_page, filter)
        tbl = self._get_entry_list("messages", limit, per_page, filter)
        # Unpack statistics
        if unpack_statistics:
            tbl.unpack_dict("statistics", prepend=False, include_original=True)
        return tbl

    def get_message(self, message_id):
        """
        `Args:`
            message_id:
               The unique id of the message
        `Returns:`
            A  JSON with the signature entry.
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/messages
        """
        return self.api.get_request(f"messages/{message_id}")

    def create_message(self, payload):
        """
        Create a message in Action Network

        `Args:`
            payload: dict
                The payload containing message details
                {
                  "subject": "Stop doing the bad thing",
                  "body": "<p>The mayor should stop doing the bad thing.</p>",
                  "from": "Progressive Action Now",
                  "reply_to": "jane@progressiveactionnow.org",
                  "targets": [
                    {
                      "href": "https://actionnetwork.org/api/v2/queries/id"
                    }
                  ],
                  "_links": {
                    "osdi:wrapper": {
                      "href": "https://actionnetwork.org/api/v2/wrappers/id"
                    }
                  }
                }
        `Returns:`
            A JSON response confirming the creation of the message
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/messages
        """
        return self.api.post_request("messages", json=payload)

    def update_message(self, message_id, payload):
        """
        Update a message in Action Network

        `Args:`
            message_id:
               The unique id of the message
            payload: dict
                The payload containing message details to be updated
                {
                    "name": "Stop doing the bad thing email send 1",
                    "subject": "Please! Stop doing the bad thing"
                }
        `Returns:`
            A JSON response confirming the update of the message
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/messages
        """
        return self.api.put_request(f"messages/{message_id}", json=payload)

    def schedule_message(self, message_id, scheduled_start_date):
        """
        Schedule a message in Action Network

        `Args:`
            message_id:
               The unique id of the message
            scheduled_start_date:
                The UTC timestamp to schedule the message at in ISO8601 format.
                e.g. "2015-03-14T12:00:00Z"
        `Returns:`
            A JSON response confirming the scheduling
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/schedule_helper
        """
        return self.api.post_request(
            f"messages/{message_id}/schedule/",
            {"scheduled_start_date": scheduled_start_date},
        )

    def send_message(self, message_id):
        """
        Send a message in Action Network

        `Args:`
            message_id:
               The unique id of the message
        `Returns:`
            A JSON response confirming the message was sent
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/send_helper
        """
        return self.api.post_request(f"messages/{message_id}/send/", {})

    # Metadata
    def get_metadata(self):
        """
        `Args:`
           None
        `Returns:`
            A  JSON with the metadata entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/metadata
        """
        return self.api.get_request("metadata")

    # Outreaches
    def get_advocacy_campaign_outreaches(
        self, advocacy_campaign_id, limit=None, per_page=25, page=None, filter=None
    ):
        """
        `Args:`
            advocacy_campaign_id:
               The unique id of the advocacy_campaign
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.

        `Returns:`
           A  JSON with all the outreaches entries related to the advocacy_campaign_id
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/outreaches
        """
        if page:
            return self._get_page(
                f"advocacy_campaigns/{advocacy_campaign_id}/outreaches",
                page,
                per_page,
                filter,
            )
        return self._get_entry_list(
            f"advocacy_campaigns/{advocacy_campaign_id}/outreaches",
            limit,
            per_page,
            filter,
        )

    def get_person_outreaches(self, person_id, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            person_id:
               The unique id of the person
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.

        `Returns:`
            A  JSON with all the outreaches entries related to our group
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/outreaches
        """
        if page:
            return self._get_page(f"people/{person_id}/outreaches", page, per_page, filter)
        return self._get_entry_list(f"people/{person_id}/outreaches", limit, per_page, filter)

    def get_advocacy_campaign_outreach(self, advocacy_campaign_id, outreach_id):
        """
        `Args:`
            advocacy_campaign_id:
               The unique id of the campaign
            outreach_id:
               The unique id of the outreach
        `Returns:`
            A  JSON with the outreach entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/outreaches
        """
        return self.api.get_request(
            f"advocacy_campaigns/{advocacy_campaign_id}/outreaches/{outreach_id}"
        )

    def get_person_outreach(self, person_id, outreach_id):
        """
        `Args:`
            person_id:
               The unique id of the campaign
            outreach_id:
               The unique id of the outreach
        `Returns:`
            A  JSON with the outreach entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/outreaches
        """
        return self.api.get_request(f"people/{person_id}/outreaches/{outreach_id}")

    def create_outreach(self, advocacy_campaign_id, payload):
        """
        Create an outreach in Action Network

        `Args:`
            advocacy_campaign_id:
                The unique id of the campaign
            payload:
                The payload containing outreach details
                {
                    "targets": [
                        {
                            "given_name": "Joe",
                            "family_name": "Schmoe"
                        }
                    ],
                    "_links" : {
                        "osdi:person" : { "href" : "https://actionnetwork.org/api/v2/people/id" }
                    }
                }
        `Returns:`
            A JSON response confirming the creation of the outreach
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/outreaches
        """
        return self.api.post_request(
            f"advocacy_campaigns/{advocacy_campaign_id}/outreaches", payload
        )

    def update_outreach(self, advocacy_campaign_id, outreach_id, payload):
        """
        Update an outreach in Action Network

        `Args:`
            advocacy_campaign_id:
                The unique id of the campaign
            outreach_id:
                The unique id of the outreach
            payload:
                The payload containing outreach details to be updated
                {
                    "subject": "Please vote no!"
                }
        `Returns:`
            A JSON response confirming the update of the outreach
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/outreaches
        """
        return self.api.put_request(
            f"advocacy_campaigns/{advocacy_campaign_id}/outreaches/{outreach_id}",
            payload,
        )

    # People
    def get_people(self, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            limit:
                The number of entries to return. When None, returns all entries.
            per_page
                The number of entries per page to return. 25 maximum.
            page
                Which page of results to return
            filter
                The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
                When None, no filter is applied.
        `Returns:`
            A list of JSONs of people stored in Action Network.
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/people
        """
        if page:
            return self._get_page("people", page, per_page, filter=filter)
        return self._get_entry_list("people", limit, per_page, filter=filter)

    def get_person(self, person_id):
        """
        `Args:`
            person_id:
                Id of the person.
        `Returns:`
            A  JSON of the entry. If the entry doesn't exist, Action Network returns
            ``{'error': 'Couldn't find person with id = <id>'}``.
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/people
        """
        return self.api.get_request(url=f"people/{person_id}")

    def upsert_person(
        self,
        email_address: Union[str, List[str], List[Dict[str, str]]] = None,
        given_name=None,
        family_name=None,
        tags=None,
        languages_spoken=None,
        postal_addresses=None,
        mobile_number=None,
        mobile_status: Literal["subscribed", "unsubscribed", None] = None,
        background_processing=False,
        **kwargs,
    ):
        """
        Creates or updates a person record. In order to update an existing record instead of
        creating a new one, you must supply an email or mobile number which matches a record
        in the database.

        Identifiers are intentionally not included as an option on
        this method, because their use can cause buggy behavior if
        they are not globally unique. ActionNetwork support strongly
        encourages developers not to use custom identifiers.

        `Args:`
            email_address:
                Either email_address or mobile_number are required. Can be any of the following
                    - a string with the person's email
                    - a list of strings with a person's emails
                    - a list of dictionaries with the following fields
                        - address (REQUIRED)
                        - primary (OPTIONAL): Boolean indicating the user's primary email address
                        - status (OPTIONAL): can taken on any of these values
                            - "subscribed"
                            - "unsubscribed"
                            - "bouncing"
                            - "previous bounce"
                            - "spam complaint"
                            - "previous spam complaint"
            given_name:
                The person's given name
            family_name:
                The person's family name
            tags:
                Optional field. A list of strings of pre-existing tags to be applied to the person.
            languages_spoken:
                Optional field. A list of strings of the languages spoken by the person
            postal_addresses:
                Optional field. A list of dictionaries.
                For details, see Action Network's documentation:
                https://actionnetwork.org/docs/v2/person_signup_helper
            mobile_number:
                Either email_address or mobile_number are required. Can be any of the following
                    - a string with the person's cell phone number
                    - an integer with the person's cell phone number
                    - a list of strings with the person's cell phone numbers
                    - a list of integers with the person's cell phone numbers
                    - a dictionary with the following fields
                        - number (REQUIRED)
                        - primary (OPTIONAL): Boolean indicating the user's primary mobile number
                        - status (OPTIONAL): can taken on any of these values
                            - "subscribed"
                            - "unsubscribed"
            mobile_status:
                None, 'subscribed' or 'unsubscribed'. If included, will update the SMS opt-in
                status of the phone in ActionNetwork. If not included, won't update the status.
                None by default, causes no updates to mobile number status. New numbers are set
                to "unsubscribed" by default.
            background_processing: bool
                If set `true`, utilize ActionNetwork's "background processing". This will return
                an immediate success, with an empty JSON body, and send your request to the
                background queue for eventual processing.
                https://actionnetwork.org/docs/v2/#background-processing
            **kwargs:
                Any additional fields to store about the person. Action Network allows
                any custom field.
        Adds a person to Action Network
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/people
        """
        email_addresses_field = None
        if isinstance(email_address, str):
            email_addresses_field = [{"address": email_address}]
        elif isinstance(email_address, list):
            if isinstance(email_address[0], str):
                email_addresses_field = [{"address": email} for email in email_address]
                email_addresses_field[0]["primary"] = True
            if isinstance(email_address[0], dict):
                email_addresses_field = email_address
        else:
            raise ValueError(
                f"Unexpected type for email_address. Got {type(email_address)}, "
                "expected str or list."
            )

        mobile_numbers_field = None
        if isinstance(mobile_number, str):
            mobile_numbers_field = [{"number": re.sub("[^0-9]", "", mobile_number)}]
        elif isinstance(mobile_number, int):
            mobile_numbers_field = [{"number": str(mobile_number)}]
        elif isinstance(mobile_number, list):
            if len(mobile_number) > 1:
                raise Exception("Action Network allows only 1 phone number per activist")
            if isinstance(mobile_number[0], list):
                mobile_numbers_field = [
                    {"number": re.sub("[^0-9]", "", cell)} for cell in mobile_number
                ]
                mobile_numbers_field[0]["primary"] = True
            if isinstance(mobile_number[0], int):
                mobile_numbers_field = [{"number": cell} for cell in mobile_number]
                mobile_numbers_field[0]["primary"] = True

        # Including status in this field changes the opt-in status in
        # ActionNetwork. This is not always desireable, so we should
        # only do so when a status is included.
        if mobile_status and mobile_numbers_field:
            for field in mobile_numbers_field:
                field["status"] = mobile_status

        # If the mobile_number field is passed a list of dictionaries, just use that directly
        if mobile_number and isinstance(mobile_number, list) and isinstance(mobile_number[0], dict):
            mobile_numbers_field = mobile_number

        if not email_addresses_field and not mobile_numbers_field:
            raise Exception(
                "Either email_address or mobile_number is required and can be formatted "
                "as a string, list of strings, a dictionary, a list of dictionaries, or "
                "(for mobile_number only) an integer or list of integers"
            )

        data = {"person": {}}

        if email_addresses_field is not None:
            data["person"]["email_addresses"] = email_addresses_field
        if mobile_numbers_field is not None:
            data["person"]["phone_numbers"] = mobile_numbers_field
        if given_name is not None:
            data["person"]["given_name"] = given_name
        if family_name is not None:
            data["person"]["family_name"] = family_name
        if languages_spoken is not None:
            data["person"]["languages_spoken"] = languages_spoken
        if postal_addresses is not None:
            data["person"]["postal_addresses"] = postal_addresses
        if tags is not None:
            data["add_tags"] = tags

        data["person"]["custom_fields"] = {**kwargs}
        url = f"{self.api_url}/people"
        if background_processing:
            url = f"{url}?background_processing=true"

        response = self.api.post_request(url, data=json.dumps(data))

        identifiers = response["identifiers"]
        person_id = [
            entry_id.split(":")[1] for entry_id in identifiers if "action_network:" in entry_id
        ]
        if not person_id:
            logger.error(f"Response gave no valid person_id: {identifiers}")
        else:
            person_id = person_id[0]
        if response["created_date"] == response["modified_date"]:
            logger.info(f"Entry {person_id} successfully added.")
        else:
            logger.info(f"Entry {person_id} successfully updated.")
        return response

    def add_person(
        self,
        email_address=None,
        given_name=None,
        family_name=None,
        tags=None,
        languages_spoken=None,
        postal_addresses=None,
        mobile_number=None,
        mobile_status="subscribed",
        **kwargs,
    ):
        """
        Creates a person in the database. WARNING: this endpoint has been deprecated in favor of
        upsert_person.
        """
        logger.warning("Method 'add_person' has been deprecated. Please use 'upsert_person'.")
        # Pass inputs to preferred method:
        self.upsert_person(
            email_address=email_address,
            given_name=given_name,
            family_name=family_name,
            languages_spoken=languages_spoken,
            postal_addresses=postal_addresses,
            mobile_number=mobile_number,
            mobile_status=mobile_status,
            **kwargs,
        )

    def update_person(self, entry_id, background_processing=False, **kwargs):
        """
        Updates a person's data in Action Network, given their Action Network ID. Note that you
        can't alter a person's tags with this method. Instead, use upsert_person.

        `Args:`
            entry_id:
                The person's Action Network id
            background_processing: bool
                If set `true`, utilize ActionNetwork's "background processing". This will return
                an immediate success, with an empty JSON body, and send your request to the
                background queue for eventual processing.
                https://actionnetwork.org/docs/v2/#background-processing
            **kwargs:
                Fields to be updated. The possible fields are
                    email_address:
                        Can be any of the following
                            - a string with the person's email
                            - a dictionary with the following fields
                                - email_address (REQUIRED)
                                    - primary (OPTIONAL): Boolean indicating the user's
                                    primary email address
                                - status (OPTIONAL): can taken on any of these values
                                    - "subscribed"
                                    - "unsubscribed"
                                    - "bouncing"
                                    - "previous bounce"
                                    - "spam complaint"
                                    - "previous spam complaint"
                    given_name:
                        The person's given name
                    family_name:
                        The person's family name
                    languages_spoken:
                        Optional field. A list of strings of the languages spoken by the person
                    postal_addresses:
                        Optional field. A list of dictionaries.
                        For details, see Action Network's documentation:
                        https://actionnetwork.org/docs/v2/people#put
                    custom_fields:
                        A dictionary of any other fields to store about the person.
                    `Documentation Reference`:
                        https://actionnetwork.org/docs/v2/people
        """
        data = {**kwargs}
        url = f"{self.api_url}/people/{entry_id}"
        if background_processing:
            url = f"{url}?background_processing=true"
        response = self.api.put_request(
            url=url,
            data=json.dumps(data),
            success_codes=[204, 201, 200],
        )
        logger.info(f"Person {entry_id} successfully updated")
        return response

    # Petitions
    def get_petitions(self, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.
        `Returns:`
            A  JSON with all of the petitions entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/petitions
        """
        if page:
            return self._get_page("petitions", page, per_page, filter)
        return self._get_entry_list("petitions", limit, per_page, filter)

    def get_petition(self, petition_id):
        """
        `Args:`
            petition_id:
               The unique id of the petition
        `Returns:`
            A  JSON with the petition entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/petitions
        """
        return self.api.get_request(f"petitions/{petition_id}")

    def create_petition(
        self, title, description, petition_text, target, background_processing=False
    ):
        """
        `Args:`
            title:
               The title of the petition
            description:
               The description of the petition
            petition_text:
               The text of the petition
            target:
               The target of the petition
            background_processing:
               Whether to process the request in the background
        `Returns:`
            A JSON with the response from the API
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/petitions
        """
        data = {
            "title": title,
            "description": description,
            "petition_text": petition_text,
            "target": target,
        }
        url = f"{self.api_url}/petitions"
        if background_processing:
            url = f"{url}?background_processing={background_processing}"
        response = self.api.post_request(
            url=url,
            data=json.dumps(data),
        )
        logger.info(f"Petition {title} successfully created")
        return response

    def update_petition(
        self,
        petition_id,
        title,
        description,
        petition_text,
        target,
        background_processing=False,
    ):
        """
        `Args:`
            petition_id:
               The unique id of the petition to be updated
            title:
               The updated title of the petition
            description:
               The updated description of the petition
            petition_text:
               The updated text of the petition
            target:
               The updated target of the petition
            background_processing:
               Whether to process the request in the background
        `Returns:`
            A JSON with the response from the API
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/petitions
        """
        data = {
            "title": title,
            "description": description,
            "petition_text": petition_text,
            "target": target,
        }
        url = f"{self.api_url}/petitions/{petition_id}"
        if background_processing:
            url = f"{url}?background_processing={background_processing}"
        response = self.api.put_request(
            url=url,
            data=json.dumps(data),
        )
        logger.info(f"Petition {title} successfully updated")
        return response

    # Queries
    def get_queries(self, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.


        `Returns:`
            A  JSON with all the query entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/queries
        """
        if page:
            return self._get_page("queries", page, per_page, filter)
        return self._get_entry_list("queries", limit, per_page, filter)

    def get_query(self, query_id):
        """
        `Args:`
           query_id:
              The unique id of the query
        `Returns:`
            A  JSON with the query entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/queries
        """
        return self.api.get_request(f"queries/{query_id}")

    # Signatures
    def get_petition_signatures(self, petition_id, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            petition_id:
               The unique id of the petition
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.
        `Returns:`
            A  JSON with all the signatures related to the petition entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/signatures
        """
        if page:
            return self._get_page(f"petitions/{petition_id}/signatures", page, per_page, filter)
        return self._get_entry_list(f"petitions/{petition_id}/signatures", limit, per_page, filter)

    def get_person_signatures(self, person_id, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            person_id:
               The unique id of the person
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.


        `Returns:`
            A  JSON with all the signatures related to the petition entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/signatures
        """
        if page:
            return self._get_page(f"people/{person_id}/signatures", page, per_page, filter)
        return self._get_entry_list(f"people/{person_id}/signatures", limit, per_page, filter)

    def get_petition_signature(self, petition_id, signature_id):
        """
        `Args:`
            petition_id:
               The unique id of the petition
            signature_id:
               The unique id of the signature
        `Returns:`
            A  JSON with the signature entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/signatures
        """
        return self.api.get_request(f"petitions/{petition_id}/signatures/{signature_id}")

    def get_person_signature(self, person_id, signature_id):
        """
        `Args:`
            person_id:
               The unique id of the person
            signature_id:
               The unique id of the signature
        `Returns:`
            A  JSON with the signature entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/signatures
        """
        return self.api.get_request(f"people/{person_id}/signatures/{signature_id}")

    def create_signature(self, petition_id, data):
        """
        `Args:`
            petition_id:
               The unique id of the petition
            data:
               The payload for creating the signature
               {
                   "comments" : "Stop doing the thing",
                   "_links" : {
                       "osdi:person" : { "href" : "https://actionnetwork.org/api/v2/people/id" }
                   }
               }
        `Returns:`
            A JSON with the created signature entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/signatures
        """
        return self.api.post_request(f"petitions/{petition_id}/signatures", data)

    def update_signature(self, petition_id, signature_id, data):
        """
        `Args:`
            petition_id:
               The unique id of the petition
            signature_id:
               The unique id of the signature
            data:
               The signature payload to update
               {
                   "comments": "Some new comments"
               }
        `Returns:`
            A JSON with the updated signature entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/signatures
        """
        return self.api.put_request(f"petitions/{petition_id}/signatures/{signature_id}", data)

    # Submissions
    def get_form_submissions(self, form_id, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            form_id:
               The unique id of the form
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.

        `Returns:`
            A  JSON with all the submissions entries related to the form
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/submissions
        """
        if page:
            return self._get_page(f"forms/{form_id}/submissions", page, per_page, filter)
        return self._get_entry_list(f"forms/{form_id}/submissions", limit, per_page, filter)

    def get_person_submissions(self, person_id, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            person_id:
               The unique id of the person
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.
        `Returns:`
            A  JSON with all the submissions entries related with our group
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/submissions
        """
        if page:
            return self._get_page(f"people/{person_id}/submissions", page, per_page, filter)
        return self._get_entry_list(f"people/{person_id}/submissions", limit, per_page, filter)

    def get_form_submission(self, form_id, submission_id):
        """
        `Args:`
            form_id:
               The unique id of the form
            submission_id:
               The unique id of the submission
        `Returns:`
            A  JSON with the submission entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/submissions
        """
        return self.api.get_request(f"forms/{form_id}/submissions/{submission_id}")

    def get_person_submission(self, person_id, submission_id):
        """
        `Args:`
            person_id:
               The unique id of the submission
            submission_id:
               The unique id of the submission
        `Returns:`
            A  JSON with the submission entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/submissions
        """
        return self.api.get_request(f"people/{person_id}/submissions/{submission_id}")

    def create_submission(self, form_id, person_id):
        """
        `Args:`
            form_id:
                The unique id of the form
            person_id:
                The unique id of the person
        `Returns:`
            A JSON response indicating the success or failure of the submission creation
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/submissions
        """
        payload = {
            "_links": {
                "osdi:person": {"href": f"https://actionnetwork.org/api/v2/people/{person_id}"}
            }
        }
        return self.api.post_request(f"forms/{form_id}/submissions", data=json.dumps(payload))

    def update_submission(self, form_id, submission_id, data):
        """
        `Args:`
            form_id:
                The unique id of the form
            submission_id:
                The unique id of the submission
            data:
                The payload for updating the submission
                {
                    "_links" : {
                        "osdi:person" : { "href" : "https://actionnetwork.org/api/v2/people/id" }
                    }
                }
        `Returns:`
            A JSON with the updated submission entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/submissions
        """
        return self.api.put_request(
            f"forms/{form_id}/submissions/{submission_id}", data=json.dumps(data)
        )

    # Tags
    def get_tags(self, limit=None, per_page=None):
        """
        `Args:`
            limit:
                The number of entries to return. When None, returns all entries.
            per_page:
                This is a deprecated argument.
        `Returns:`
            A list of JSONs of tags in Action Network.
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/tags
        """
        if per_page:
            warnings.warn(
                "per_page is a deprecated argument on get_tags()",
                DeprecationWarning,
                stacklevel=2,
            )
        return self._get_entry_list("tags", limit)

    def get_tag(self, tag_id):
        """
        `Args:`
            tag_id:
                Id of the tag.
        `Returns:`
            A  JSON of the entry. If the entry doesn't exist, Action Network returns
            "{'error': 'Couldn't find tag with id = <id>'}"
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/tags
        """
        return self.api.get_request(url=f"tags/{tag_id}")

    def add_tag(self, name):
        """
        `Args:`
            name:
                The tag's name. This is the ONLY editable field
        Adds a tag to Action Network. Once created, tags CANNOT be edited or deleted.
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/tags
        """
        data = {"name": name}
        response = self.api.post_request(url=f"{self.api_url}/tags", data=json.dumps(data))
        identifiers = response["identifiers"]
        person_id = [
            entry_id.split(":")[1] for entry_id in identifiers if "action_network:" in entry_id
        ][0]
        logger.info(f"Tag {person_id} successfully added to tags.")
        return response

    # Taggings
    def get_taggings(self, tag_id, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            tag_id:
                The unique id of the tag
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.


        `Returns:`
            A  JSON with all the tagging entries associated with the tag_id
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/taggings
        """
        if page:
            return self._get_page(f"tags/{tag_id}/taggings", page, per_page, filter)
        return self._get_entry_list(f"tags/{tag_id}/taggings", limit, per_page, filter)

    def get_tagging(self, tag_id, tagging_id):
        """
        `Args:`
           tag_id:
              The unique id of the tag
           tagging_id:
              The unique id of the tagging
        `Returns:`
            A  JSON with the tagging entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/taggings
        """
        return self.api.get_request(f"tags/{tag_id}/taggings/{tagging_id}")

    def create_tagging(self, tag_id, payload, background_processing=False):
        """
        `Args:`
            tag_id:
                The unique id of the tag
            payload:
                The payload for creating the tagging
                {
                    "_links" : {
                        "osdi:person" : { "href" : "https://actionnetwork.org/api/v2/people/id" }
                    }
                }
            background_processing: bool
                If set `true`, utilize ActionNetwork's "background processing". This will return
                an immediate success, with an empty JSON body, and send your request to the
                background queue for eventual processing.
                https://actionnetwork.org/docs/v2/#background-processing
        `Returns:`
            A JSON response after creating the tagging
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/taggings
        """
        url = f"tags/{tag_id}/taggings"
        if background_processing:
            url = f"{url}?background_processing=true"
        return self.api.post_request(url, data=json.dumps(payload))

    def delete_tagging(self, tag_id, tagging_id, background_processing=False):
        """
        `Args:`
            tag_id:
                The unique id of the tag
            tagging_id:
                The unique id of the tagging to be deleted
            background_processing: bool
                If set `true`, utilize ActionNetwork's "background processing". This will return
                an immediate success, with an empty JSON body, and send your request to the
                background queue for eventual processing.
                https://actionnetwork.org/docs/v2/#background-processing
        `Returns:`
            A JSON response after deleting the tagging
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/taggings
        """
        url = f"tags/{tag_id}/taggings/{tagging_id}"
        if background_processing:
            url = f"{url}?background_processing=true"
        return self.api.delete_request(url)

    # Wrappers
    def get_wrappers(self, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            limit:
               The number of entries to return. When None, returns all entries.
           per_page:
               The number of entries per page to return. 25 maximum.
           page:
               Which page of results to return
           filter:
               The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
               When None, no filter is applied.


        `Returns:`
            A  JSON with all the wrapper entries
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/wrappers
        """
        if page:
            return self._get_page("wrappers", page, per_page, filter)
        return self._get_entry_list("wrappers", limit, per_page, filter)

    def get_wrapper(self, wrapper_id):
        """
        `Args:`
           wrapper_id:
              The unique id of the wrapper
           tagging_id:
              The unique id of the tagging
        `Returns:`
            A  JSON with the wrapper entry
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/wrappers
        """
        return self.api.get_request(f"wrappers/{wrapper_id}")

    # Unique ID Lists
    def get_unique_id_lists(self, limit=None, per_page=25, page=None, filter=None):
        """
        Args:
            limit: The maximum number of unique ID lists to return.
            When None, returns all unique ID lists.
            per_page: The number of unique ID lists to return per page. Default is 25.
            page: The specific page of unique ID lists to return.
            filter: The filter criteria to apply when retrieving unique ID lists.
        Returns:
            A JSON response with the unique ID lists.
        Documentation Reference:
            https://actionnetwork.org/docs/v2/unique_id_lists
        """
        if page:
            return self._get_page("unique_id_lists", page, per_page, filter)
        return self._get_entry_list("unique_id_lists", limit, per_page, filter)

    def get_unique_id_list(self, unique_id_list_id):
        """
        `Args:`
            unique_id_list_id:
                The unique id of the unique ID list
        `Returns:`
            A JSON response with the unique ID list details
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/unique_id_lists
        """
        return self.api.get_request(f"unique_id_lists/{unique_id_list_id}")

    def create_unique_id_list(self, list_name, unique_ids):
        """
        `Args:`
            list_name:
                The name for the new list
            unique_ids:
                An array of unique IDs to upload
        `Returns:`
            A JSON response with the unique ID list details
        `Documentation Reference`:
            https://actionnetwork.org/docs/v2/unique_id_lists
        """
        return self.api.post_request(
            "unique_id_lists",
            data=json.dumps({"name": list_name, "unique_ids": unique_ids}),
        )



import json
import os
import unittest

import requests_mock
from slackclient.exceptions import SlackClientError

from parsons import Slack, Table

_dir = os.path.dirname(__file__)
responses_dir = f"{_dir}/responses"


class TestSlack(unittest.TestCase):
    def setUp(self):
        os.environ["SLACK_API_TOKEN"] = "SOME_API_TOKEN"

        self.slack = Slack()

    def tearDown(self):
        pass

    def test_slack_init(self):
        # Delete to test that is raises an error
        del os.environ["SLACK_API_TOKEN"]

        self.assertNotIn("SLACK_API_TOKEN", os.environ)

        self.assertRaises(KeyError, Slack)

        os.environ["SLACK_API_TOKEN"] = "SOME_API_TOKEN"
        self.assertIn("SLACK_API_TOKEN", os.environ)

    @requests_mock.Mocker()
    def test_channels(self, m):
        with open(f"{responses_dir}/channels.json", "r") as f:
            slack_resp = json.load(f)

        m.post("https://slack.com/api/conversations.list", json=slack_resp)

        tbl = self.slack.channels()

        self.assertIsInstance(tbl, Table)

        expected_columns = ["id", "name"]
        self.assertListEqual(tbl.columns, expected_columns)
        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_channels_all_fields(self, m):
        with open(f"{responses_dir}/channels.json", "r") as f:
            slack_resp = json.load(f)

        m.post("https://slack.com/api/conversations.list", json=slack_resp)

        fields_req = [
            "id",
            "name",
            "is_channel",
            "created",
            "creator",
            "is_archived",
            "is_general",
            "name_normalized",
            "is_shared",
            "is_org_shared",
            "is_member",
            "is_private",
            "is_mpim",
            "members",
            "topic_value",
            "topic_creator",
            "topic_last_set",
            "purpose_value",
            "purpose_creator",
            "purpose_last_set",
            "previous_names",
            "num_members",
        ]
        tbl = self.slack.channels(fields=fields_req)

        self.assertIsInstance(tbl, Table)

        expected_columns = [
            "id",
            "name",
            "is_channel",
            "created",
            "creator",
            "is_archived",
            "is_general",
            "name_normalized",
            "is_shared",
            "is_org_shared",
            "is_member",
            "is_private",
            "is_mpim",
            "members",
            "topic_value",
            "topic_creator",
            "topic_last_set",
            "purpose_value",
            "purpose_creator",
            "purpose_last_set",
            "previous_names",
            "num_members",
        ]

        self.assertListEqual(sorted(tbl.columns), sorted(expected_columns))
        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_users(self, m):
        with open(f"{responses_dir}/users.json", "r") as f:
            slack_resp = json.load(f)

        m.post("https://slack.com/api/users.list", json=slack_resp)

        tbl = self.slack.users()

        self.assertIsInstance(tbl, Table)

        expected_columns = [
            "id",
            "name",
            "deleted",
            "profile_email",
            "profile_real_name_normalized",
        ]
        self.assertListEqual(tbl.columns, expected_columns)
        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_users_all_fields(self, m):
        with open(f"{responses_dir}/users.json", "r") as f:
            slack_resp = json.load(f)

        m.post("https://slack.com/api/users.list", json=slack_resp)

        fields_req = [
            "id",
            "team_id",
            "name",
            "deleted",
            "color",
            "real_name",
            "tz",
            "tz_label",
            "tz_offset",
            "is_admin",
            "is_owner",
            "is_primary_owner",
            "is_restricted",
            "is_ultra_restricted",
            "is_bot",
            "updated",
            "is_app_user",
            "has_2fa",
            "profile_avatar_hash",
            "profile_display_name",
            "profile_display_name_normalized",
            "profile_email",
            "profile_first_name",
            "profile_image_1024",
            "profile_image_192",
            "profile_image_24",
            "profile_image_32",
            "profile_image_48",
            "profile_image_512",
            "profile_image_72",
            "profile_image_original",
            "profile_last_name",
            "profile_phone",
            "profile_real_name",
            "profile_real_name_normalized",
            "profile_skype",
            "profile_status_emoji",
            "profile_status_text",
            "profile_team",
            "profile_title",
        ]
        tbl = self.slack.users(fields=fields_req)

        self.assertIsInstance(tbl, Table)

        expected_columns = [
            "id",
            "team_id",
            "name",
            "deleted",
            "color",
            "real_name",
            "tz",
            "tz_label",
            "tz_offset",
            "is_admin",
            "is_owner",
            "is_primary_owner",
            "is_restricted",
            "is_ultra_restricted",
            "is_bot",
            "updated",
            "is_app_user",
            "has_2fa",
            "profile_avatar_hash",
            "profile_display_name",
            "profile_display_name_normalized",
            "profile_email",
            "profile_first_name",
            "profile_image_1024",
            "profile_image_192",
            "profile_image_24",
            "profile_image_32",
            "profile_image_48",
            "profile_image_512",
            "profile_image_72",
            "profile_image_original",
            "profile_last_name",
            "profile_phone",
            "profile_real_name",
            "profile_real_name_normalized",
            "profile_skype",
            "profile_status_emoji",
            "profile_status_text",
            "profile_team",
            "profile_title",
        ]
        self.assertListEqual(sorted(tbl.columns), sorted(expected_columns))
        self.assertEqual(tbl.num_rows, 2)

    @requests_mock.Mocker()
    def test_message_channel(self, m):
        with open(f"{responses_dir}/message_channel.json", "r") as f:
            slack_resp = json.load(f)

        m.post("https://slack.com/api/chat.postMessage", json=slack_resp)

        dct = self.slack.message_channel("C1H9RESGL", "Here's a message for you")

        self.assertIsInstance(dct, dict)
        self.assertListEqual(sorted(dct), sorted(slack_resp))

        m.post(
            "https://slack.com/api/chat.postMessage",
            json={"ok": False, "error": "invalid_auth"},
        )

        self.assertRaises(
            SlackClientError,
            self.slack.message_channel,
            "FakeChannel",
            "Here's a message for you",
        )

    @requests_mock.Mocker(case_sensitive=True)
    def test_message(self, m):
        webhook = "https://hooks.slack.com/services/T1234/B1234/D12322"
        m.post(webhook, json={"ok": True})
        Slack.message("#foobar", "this is a message", webhook)
        self.assertEqual(
            m._adapter.last_request.json(),
            {"text": "this is a message", "channel": "#foobar"},
        )
        self.assertEqual(m._adapter.last_request.path, "/services/T1234/B1234/D12322")

    @requests_mock.Mocker()
    def test_file_upload(self, m):
        file_path = f"{responses_dir}/file_upload.json"
        with open(file_path, "r") as f:
            slack_resp = json.load(f)

        m.post("https://slack.com/api/files.upload", json=slack_resp)

        dct = self.slack.upload_file(["D0L4B9P0Q"], file_path)

        self.assertIsInstance(dct, dict)
        self.assertListEqual(sorted(dct), sorted(slack_resp))

        m.post(
            "https://slack.com/api/files.upload",
            json={"ok": False, "error": "invalid_auth"},
        )

        self.assertRaises(SlackClientError, self.slack.upload_file, ["D0L4B9P0Q"], file_path)



folder_json = {
    "folders": [
        {
            "id": "123",
            "name": "Voter Events",
            "parent": "0",
            "permissions": "full",
            "subfolders": [
                {
                    "id": "200",
                    "name": "Phone Banks",
                    "parent": "123",
                    "permissions": "full",
                },
                {
                    "id": "201",
                    "name": "Text Banks",
                    "parent": "121",
                    "permissions": "full",
                },
            ],
        },
        {"id": "121", "name": "Fundraising", "parent": "0", "permissions": "full"},
        {"id": "122", "name": "Internal", "parent": "0", "permissions": "full"},
    ],
    "total": 3,
}
form_json = {
    "forms": [
        {
            "id": "1001",
            "created": "2018-08-01 10:10:00",
            "db": "1",
            "deleted": "0",
            "folder": "123",
            "language": "en",
            "name": "Test Form 1",
            "num_columns": "3",
            "progress_meter": "0",
            "submissions": "1321",
            "submissions_unread": "1",
            "updated": "2019-10-10 11:04:53",
            "viewkey": "viewkey_test",
            "views": "1000",
            "submissions_today": 0,
            "last_submission_id": "54321",
            "last_submission_time": "2020-10-10 10:10:10",
            "url": "https://test.formstack.com/forms/test_1",
            "encrypted": False,
            "thumbnail_url": None,
            "submit_button_title": "Submit",
            "inactive": True,
            "timezone": "US/Eastern",
            "should_display_one_question_at_a_time": False,
            "can_access_1q_feature": True,
            "is_workflow_form": False,
            "is_workflow_published": False,
            "has_approvers": False,
            "edit_url": "https://www.formstack.com/admin/form/builder/1001/build",
            "data_url": "",
            "summary_url": "",
            "rss_url": "",
            "permissions": 150,
            "can_edit": True,
        },
        {
            "id": "1002",
            "created": "2018-08-01 10:10:00",
            "db": "1",
            "deleted": "0",
            "folder": "123",
            "language": "en",
            "name": "Test Form 2",
            "num_columns": "3",
            "progress_meter": "0",
            "submissions": "1321",
            "submissions_unread": "1",
            "updated": "2019-10-10 11:04:53",
            "viewkey": "viewkey_test",
            "views": "1000",
            "submissions_today": 0,
            "last_submission_id": "54321",
            "last_submission_time": "2020-10-10 10:10:10",
            "url": "https://test.formstack.com/forms/test_2",
            "encrypted": False,
            "thumbnail_url": None,
            "submit_button_title": "Submit",
            "inactive": True,
            "timezone": "US/Eastern",
            "should_display_one_question_at_a_time": False,
            "can_access_1q_feature": True,
            "is_workflow_form": False,
            "is_workflow_published": False,
            "has_approvers": False,
            "edit_url": "https://www.formstack.com/admin/form/builder/1002/build",
            "data_url": "",
            "summary_url": "",
            "rss_url": "",
            "permissions": 150,
            "can_edit": True,
        },
    ],
    "total": 3,
}
submission_id = 332525567
submission_json = {
    "id": str(submission_id),
    "timestamp": "2017-06-13 17:07:51",
    "user_agent": "Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko",
    "remote_addr": "50.232.88.142",
    "payment_status": "",
    "form": "2691007",
    "latitude": "39.955799102783",
    "longitude": "-105.16829681396",
    "data": [
        {"field": "52404588", "value": "95386.00000"},
        {"field": "52404587", "value": "lisetalbott@gmail.com"},
    ],
    "pretty_field_id": "52404587",
}
form_submissions_json = {
    "submissions": [
        {
            "id": "325343738",
            "timestamp": "2017-05-07 21:52:42",
            "user_agent": (
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_4) AppleWebKit/537.36"
                " (KHTML, like Gecko) Chrome/57.0.2987.133 Safari/537.36"
            ),
            "remote_addr": "73.163.190.175",
            "payment_status": "",
            "latitude": "38.905700683594",
            "longitude": "-76.978302001953",
            "read": "1",
        },
        {
            "id": "325512020",
            "timestamp": "2017-05-08 15:31:08",
            "user_agent": (
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/601.7.8"
                " (KHTML, like Gecko) Version/9.1.3 Safari/601.7.8"
            ),
            "remote_addr": "65.242.21.250",
            "payment_status": "",
            "latitude": "37.750999450684",
            "longitude": "-97.821998596191",
            "read": "1",
        },
        {
            "id": "325512197",
            "timestamp": "2017-05-08 15:31:52",
            "user_agent": (
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_4) AppleWebKit/603.1.30"
                " (KHTML, like Gecko) Version/10.1 Safari/603.1.30"
            ),
            "remote_addr": "65.242.21.250",
            "payment_status": "",
            "latitude": "37.750999450684",
            "longitude": "-97.821998596191",
            "read": "1",
        },
    ],
    "total": 1321,
    "pages": 53,
    "pretty_field_id": "52404587",
}
form_fields_json = [
    {
        "id": "52404588",
        "label": "Zip",
        "hide_label": "0",
        "description": "",
        "name": "zip",
        "type": "number",
        "options": "",
        "required": "1",
        "uniq": "0",
        "hidden": "0",
        "readonly": "0",
        "colspan": "1",
        "sort": "0",
        "logic": None,
        "calculation": "",
        "workflow_access": "write",
        "default": "",
        "field_one_calculation": 0,
        "field_two_calculation": 0,
        "calculation_units": "",
        "calculation_operator": "",
        "calculation_type": "",
        "calculation_category": "",
        "calculation_allow_negatives": 0,
        "text_size": 5,
        "min_value": "",
        "max_value": "",
        "currency": "",
        "decimals": "5",
        "slider": "0",
        "placeholder": "",
    },
    {
        "id": "52404587",
        "label": "Email",
        "hide_label": "0",
        "description": "",
        "name": "email",
        "type": "email",
        "options": "",
        "required": "1",
        "uniq": "0",
        "hidden": "0",
        "readonly": "0",
        "colspan": "1",
        "sort": "1",
        "logic": None,
        "calculation": "",
        "workflow_access": "write",
        "default": "",
        "text_size": 50,
        "maxlength": 0,
        "confirm": 0,
        "confirmationText": "",
        "placeholder": "",
    },
]



import datetime
import decimal
import os
import unittest

import requests_mock

from parsons import Braintree, Table
from test.utils import assert_matching_tables

_dir = os.path.dirname(__file__)


class TestBraintree(unittest.TestCase):
    def setUp(self):
        self.braintree = Braintree(
            merchant_id="abcd1234abcd1234",
            public_key="abcd1234abcd1234",
            private_key="abcd1234abcd1234abcd1234abcd1234",
        )

    @requests_mock.Mocker()
    def test_dispute_search(self, m):
        m.post(
            "https://api.braintreegateway.com:443"
            "/merchants/abcd1234abcd1234/disputes/advanced_search?page=1",
            text=open(f"{_dir}/test_data/dispute_example.xml").read(),
        )
        table = self.braintree.get_disputes(start_date="2020-01-01", end_date="2020-01-02")

        self.assertEqual(len(table.table), 3)
        self.assertEqual(table[0]["id"], "abcd1234abcd1234")
        self.assertEqual(table[1]["id"], "ghjk6789ghjk6789")
        self.assertEqual(table[0]["transaction_id"], "d9f876fg")
        self.assertEqual(table[1]["transaction_id"], "98df87fg")
        self.assertEqual(table[0]["reason"], "transaction_amount_differs")
        self.assertEqual(table[1]["reason"], "fraud")

    @requests_mock.Mocker()
    def test_transaction_search(self, m):
        m.post(
            "https://api.braintreegateway.com:443"
            "/merchants/abcd1234abcd1234/transactions/advanced_search_ids",
            text="""
               <search-results>
                  <page-size type="integer">50</page-size>
                  <ids type="array"><item>1234abcd</item> <item>0987asdf</item> </ids>
               </search-results>
        """,
        )
        table = self.braintree.get_transactions(
            disbursement_start_date="2020-01-01",
            disbursement_end_date="2020-01-02",
            just_ids=True,
        )
        assert_matching_tables(table, Table([["id"], ["1234abcd"], ["0987asdf"]]))
        m.post(
            "https://api.braintreegateway.com:443"
            "/merchants/abcd1234abcd1234/transactions/advanced_search",
            text=open(f"{_dir}/test_data/transaction_example.xml").read(),
        )
        full_table = self.braintree.get_transactions(
            disbursement_start_date="2020-01-01",
            disbursement_end_date="2020-01-02",
            table_of_ids=table,
        )
        self.assertEqual(len(table.table), 3)
        self.assertEqual(len(full_table.table), 3)
        self.assertEqual(table[0]["id"], "1234abcd")
        self.assertEqual(table[1]["id"], "0987asdf")
        self.assertEqual(len(table[0].keys()), 1)
        self.assertEqual(len(full_table[0].keys()), 67)

        self.assertEqual(full_table[0]["disbursement_date"], datetime.date(2019, 12, 30))
        self.assertEqual(full_table[0]["credit_card_bin"], "789234")
        self.assertEqual(full_table[0]["disbursement_success"], True)
        self.assertEqual(full_table[0]["amount"], decimal.Decimal("150.00"))

    @requests_mock.Mocker()
    def test_subscription_search(self, m):
        m.post(
            "https://api.braintreegateway.com:443"
            "/merchants/abcd1234abcd1234/subscriptions/advanced_search_ids",
            text="""
               <search-results>
                  <page-size type="integer">50</page-size>
                  <ids type="array"><item>aabbcc</item> <item>1a2b3c</item> </ids>
               </search-results>
        """,
        )
        table = self.braintree.get_subscriptions(
            start_date="2022-08-22", end_date="2022-08-23", just_ids=True
        )
        assert_matching_tables(table, Table([["id"], ["aabbcc"], ["1a2b3c"]]))
        m.post(
            "https://api.braintreegateway.com:443"
            "/merchants/abcd1234abcd1234/subscriptions/advanced_search",
            text=open(f"{_dir}/test_data/subscription_example.xml").read(),
        )
        full_table = self.braintree.get_subscriptions(
            start_date="2020-01-01",
            end_date="2020-01-02",
            table_of_ids=table,
            include_transactions=True,
        )
        self.assertEqual(len(table.table), 3)
        self.assertEqual(len(full_table.table), 3)
        self.assertEqual(table[0]["id"], "aabbcc")
        self.assertEqual(table[1]["id"], "1a2b3c")
        self.assertEqual(len(table[0].keys()), 1)
        self.assertEqual(len(full_table[0].keys()), 33)

        self.assertEqual(full_table[0]["first_billing_date"], datetime.date(2022, 8, 22))
        self.assertEqual(full_table[0]["transactions"][0].credit_card_details.bin, "999")
        self.assertEqual(full_table[0]["never_expires"], True)
        self.assertEqual(full_table[0]["price"], decimal.Decimal("10.00"))

    def test_query_generation(self):
        query = self.braintree._get_query_objects(
            "transaction",
            **{"disbursement_date": {"between": ["2020-01-01", "2020-01-01"]}},
        )
        self.assertEqual(query[0].name, "disbursement_date")
        self.assertEqual(query[0].to_param(), {"min": "2020-01-01", "max": "2020-01-01"})

        query = self.braintree._get_query_objects(
            "transaction", **{"merchant_account_id": {"in_list": ["abc123"]}}
        )

        self.assertEqual(query[0].name, "merchant_account_id")
        self.assertEqual(query[0].to_param(), ["abc123"])

        query = self.braintree._get_query_objects(
            "dispute",
            **{
                "merchant_account_id": {"in_list": ["abc123"]},
                "effective_date": {"between": ["2020-01-01", "2020-01-01"]},
            },
        )
        self.assertEqual(query[0].name, "merchant_account_id")
        self.assertEqual(query[1].name, "effective_date")
        self.assertEqual(query[1].to_param(), {"min": "2020-01-01", "max": "2020-01-01"})



import os
import unittest
import unittest.mock as mock

import requests_mock

from parsons import VAN, Table
from parsons.utilities import cloud_storage
from test.utils import validate_list

os.environ["VAN_API_KEY"] = "SOME_KEY"


class TestScores(unittest.TestCase):
    def setUp(self):
        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters")

    @requests_mock.Mocker()
    def test_get_scores(self, m):
        json = {
            "count": 2,
            "items": [
                {
                    "origin": None,
                    "scoreId": 2716,
                    "name": "Democratic Party Support",
                    "maxValue": 100.0,
                    "minValue": 1.0,
                    "state": None,
                    "shortName": "Dem Support",
                    "description": None,
                }
            ],
            "nextPageLink": None,
        }

        m.get(self.van.connection.uri + "scores", json=json)

        expected = [
            "origin",
            "scoreId",
            "name",
            "maxValue",
            "minValue",
            "state",
            "shortName",
            "description",
        ]

        self.assertTrue(validate_list(expected, self.van.get_scores()))

    @requests_mock.Mocker()
    def test_get_score(self, m):
        score_id = 2716

        json = {
            "origin": None,
            "scoreId": 2716,
            "name": "Democratic Party Support",
            "maxValue": 100.0,
            "minValue": 1.0,
            "state": None,
            "shortName": "Dem Support",
            "description": None,
        }

        m.get(self.van.connection.uri + "scores/{}".format(score_id), json=json)
        self.assertEqual(json, self.van.get_score(score_id))

    @requests_mock.Mocker()
    def test_get_score_updates(self, m):
        json = {
            "items": [
                {
                    "scoreUpdateId": 58319,
                    "score": {
                        "scoreId": 29817,
                        "name": "TargetSmart Gun Ownership",
                        "shortName": None,
                        "description": None,
                        "minValue": 0.0,
                        "maxValue": 100.0,
                        "state": "MT",
                        "origin": None,
                    },
                    "updateStatistics": {
                        "totalRows": 856644,
                        "duplicateRows": 0,
                        "matchedRows": 856644,
                        "matchPercent": 100.0,
                        "increasedBy": 441264,
                        "decreasedBy": 280588,
                        "nulledOut": 3649,
                        "added": 115129,
                        "outOfRange": 0,
                        "badValues": 0,
                        "maxValue": 95.9,
                        "minValue": 11.2,
                        "averageValue": 72.3338,
                        "medianValue": 76.3,
                    },
                    "loadStatus": "Completed",
                    "dateProcessed": "2019-09-10T02:07:00Z",
                }
            ],
            "nextPageLink": None,
            "count": 306,
        }

        m.get(self.van.connection.uri + "scoreUpdates", json=json)

        expected = [
            "scoreUpdateId",
            "loadStatus",
            "dateProcessed",
            "added",
            "averageValue",
            "badValues",
            "decreasedBy",
            "duplicateRows",
            "increasedBy",
            "matchPercent",
            "matchedRows",
            "maxValue",
            "medianValue",
            "minValue",
            "nulledOut",
            "outOfRange",
            "totalRows",
            "description",
            "maxValue",
            "minValue",
            "name",
            "origin",
            "scoreId",
            "shortName",
            "state",
        ]

        self.assertTrue(validate_list(expected, self.van.get_score_updates()))

    @requests_mock.Mocker()
    def test_get_score_update(self, m):
        score_update_id = 27892

        json = {
            "loadStatus": "Canceled",
            "updateStatistics": {
                "increasedBy": 1,
                "nulledOut": 1,
                "added": 0,
                "matchedRows": 4,
                "matchPercent": 100.0,
                "outOfRange": 0,
                "badValues": 1,
                "totalRows": 4,
                "maxValue": 30.0,
                "medianValue": 15.0,
                "minValue": 10.0,
                "duplicateRows": "null",
                "averageValue": 20.0,
                "decreasedBy": 2,
            },
            "score": {
                "origin": "null",
                "scoreId": 2716,
                "name": "Democratic Party Support",
                "maxValue": 100.0,
                "minValue": 1.0,
                "state": "null",
                "shortName": "null",
                "description": "null",
            },
            "dateProcessed": "null",
            "scoreUpdateId": 27892,
        }

        m.get(self.van.connection.uri + f"scoreUpdates/{score_update_id}", json=json)

        # expected = ['loadStatus', 'updateStatistics', 'score', 'dateProcessed', 'scoreUpdateId']

        self.assertEqual(json, self.van.get_score_update(score_update_id))

    @requests_mock.Mocker()
    def test_update_score_status(self, m):
        score_update_id = 27892

        m.patch(
            self.van.connection.uri + "scoreUpdates/{}".format(score_update_id),
            status_code=204,
        )

        # Test bad input
        self.assertRaises(ValueError, self.van.update_score_status, score_update_id, "not a thing.")

        # Test good input
        self.assertTrue(self.van.update_score_status(score_update_id, "approved"))

    @requests_mock.Mocker()
    def test_upload_scores(self, m):
        # Mock Cloud Storage
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = "https://box.com/my_file.zip"

        # Test uploading a job
        tbl = Table([["vanid", "col"], ["1", ".5"]])
        json = {"jobId": 9749}
        m.post(self.van.connection.uri + "FileLoadingJobs", json=json, status_code=201)
        self.van.upload_scores(tbl, [{"score_id": 9999, "score_column": "col"}], url_type="S3")

    @requests_mock.Mocker()
    def test_create_file_load(self, m):
        file_name = "test_scores.csv"
        file_url_good = "http://tmc.org/test_scores.zip"
        # file_url_bad = 'http://tmc.org/test_scores'
        columns = ["vanid", "score"]
        id_column = "vanid"
        id_type = "VANID"
        score_id = 2716
        score_column = "score"
        bad_delimiter = "*"

        json = {"jobId": 9749}

        m.post(self.van.connection.uri + "FileLoadingJobs", json=json, status_code=201)

        # Test bad delimiter
        self.assertRaises(
            ValueError,
            self.van.create_file_load,
            file_name,
            file_url_good,
            columns,
            id_column,
            id_type,
            score_id,
            score_column,
            delimiter=bad_delimiter,
        )

        # Test good request
        self.assertEqual(
            json["jobId"],
            self.van.create_file_load(
                file_name,
                file_url_good,
                columns,
                id_column,
                id_type,
                score_id,
                score_column,
            ),
        )



import logging
from typing import Optional, Union

from parsons.databases.database_connector import DatabaseConnector
from parsons.etl.table import Table

logger = logging.getLogger(__name__)


class FakeDatabase(DatabaseConnector):
    def __init__(self):
        self.table_map = {}
        self.copy_call_args = []

    def query(self, sql: str, parameters: Optional[Union[list, dict]] = None) -> Table:
        return Table()

    def table_exists(self, table_name: str) -> bool:
        return table_name in self.table_map

    def setup_table(self, table_name, data, failures=0):
        self.table_map[table_name] = {
            "failures": failures,
            "table": FakeTable(table_name, data),
        }
        return self.table_map[table_name]["table"]

    def table(self, table_name):
        if table_name not in self.table_map:
            self.setup_table(table_name, None)

        return self.table_map[table_name]["table"]

    def copy(self, data, table_name, **kwargs):
        logger.info("Copying %s rows", data.num_rows)
        if table_name not in self.table_map:
            self.setup_table(table_name, Table())

        if self.table_map[table_name]["table"].data is None:
            self.table_map[table_name]["table"].data = Table()

        if self.table_map[table_name]["failures"] > 0:
            self.table_map[table_name]["failures"] -= 1
            raise ValueError("Canned error")

        self.copy_call_args.append(
            {
                "data": data,
                "table_name": table_name,
                "kwargs": kwargs,
            }
        )

        tbl = self.table_map[table_name]["table"]

        tbl.data.concat(data)

    def get_table_object(self, table_name):
        pass

    def create_table(self, table_object, table_name):
        pass


class FakeTable:
    def __init__(self, table_name, data):
        self.table_name = table_name
        self.data = data

    def drop(self, cascade=False):
        self.data = None

    def truncate(self):
        self.data = Table()

    def distinct_primary_key(self, primary_key):
        if primary_key not in self.data.columns:
            return True

        pk_values = list(self.data[primary_key])
        pk_set = set(pk_values)
        return len(pk_set) == len(pk_values)

    @property
    def columns(self):
        return self.data.columns

    def max_primary_key(self, primary_key):
        if primary_key not in self.data.columns:
            return None

        return max(self.data[primary_key])

    @property
    def num_rows(self):
        return self.data.num_rows

    @property
    def exists(self):
        return self.data is not None

    def get_rows(self, offset=0, chunk_size=None, order_by=None):
        if not self.data:
            return self.data

        data = self.data.cut(*self.data.columns)

        if order_by:
            data.sort(order_by)

        if chunk_size:
            subset = data[offset : chunk_size + offset]
        else:
            subset = data[offset:]

        return Table(subset)

    def get_new_rows_count(self, primary_key_col, start_value=None):
        data = self.data.select_rows(lambda row: row[primary_key_col] > start_value)
        return data.num_rows

    def get_new_rows(self, primary_key, cutoff_value, offset=0, chunk_size=None):
        data = self.data.select_rows(lambda row: row[primary_key] > cutoff_value)
        data.sort(primary_key)

        return Table(data[offset : chunk_size + offset])



import json
import logging
import math
import time

import requests

from parsons.etl.table import Table
from parsons.utilities import check_env

logger = logging.getLogger(__name__)


class ActionKit(object):
    """
    Instantiate the ActionKit class

    `Args:`
        domain: str
            The ActionKit domain (e.g. ``myorg.actionkit.com``) Not required if
            ``ACTION_KIT_DOMAIN`` env variable set.
        username: str
            The authorized ActionKit username. Not required if ``ACTION_KIT_USERNAME`` env
            variable set.
        password: str
            The authorized ActionKit user password. Not required if ``ACTION_KIT_PASSWORD``
            env variable set.
    """

    _default_headers = {
        "content-type": "application/json",
        "accepts": "application/json",
    }

    def __init__(self, domain=None, username=None, password=None):
        self.domain = check_env.check("ACTION_KIT_DOMAIN", domain)
        self.username = check_env.check("ACTION_KIT_USERNAME", username)
        self.password = check_env.check("ACTION_KIT_PASSWORD", password)
        self.conn = self._conn()

    def _conn(self, default_headers=_default_headers):
        client = requests.Session()
        client.auth = (self.username, self.password)
        client.headers.update(default_headers)
        return client

    def _base_endpoint(self, endpoint, entity_id=None):
        # Create the base endpoint URL

        url = f"https://{self.domain}/rest/v1/{endpoint}/"

        if entity_id:
            return url + f"{entity_id}/"
        return url

    def _base_get(self, endpoint, entity_id=None, exception_message=None, params=None):
        # Make a general get request to ActionKit

        resp = self.conn.get(self._base_endpoint(endpoint, entity_id), params=params)
        if exception_message and resp.status_code == 404:
            raise Exception(self.parse_error(resp, exception_message))
        return resp.json()

    def _base_post(self, endpoint, exception_message, return_full_json=False, **kwargs):
        # Make a general post request to ActionKit

        resp = self.conn.post(self._base_endpoint(endpoint), data=json.dumps(kwargs))

        if resp.status_code != 201:
            raise Exception(self.parse_error(resp, exception_message))

        # Some of the methods should just return pointer to location of created
        # object.
        if "headers" in resp.__dict__ and not return_full_json:
            return resp.__dict__["headers"]["Location"]

        # Not all responses return a json
        try:
            return resp.json()

        except ValueError:
            return None

    def parse_error(self, resp, exception_message):
        # AK provides some pretty robust/helpful error reporting. We should surface them with
        # our exceptions.

        if "errors" in resp.json().keys():
            if isinstance(resp.json()["errors"], list):
                exception_message += "\n" + ",".join(resp.json()["errors"])
            else:
                for k, v in resp.json()["errors"].items():
                    exception_message += str("\n" + k + ": " + ",".join(v))

        return exception_message

    def get_user(self, user_id):
        """
        Get a user.

        `Args:`
            user_id: int
                The user id of the record to get.
        `Returns`:
            User json object
        """

        return self._base_get(
            endpoint="user", entity_id=user_id, exception_message="User not found"
        )

    def get_user_fields(self):
        """
        Get list of valid user fields that can be passed with the
        :meth:`ActionKit.create_user` method.

        `Returns`:
            List of user fields
        """

        resp = self._base_get(endpoint="user/schema")

        return list(resp["fields"].keys())

    def create_user(self, email, **kwargs):
        """
        Create a user.

        `Args:`
            email: str
                Email for the user
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns:`
            User json object
        """

        return self._base_post(
            endpoint="user",
            exception_message="Could not create user",
            email=email,
            **kwargs,
        )

    def add_phone(self, user_id, phone_type, phone):
        """
        Add a phone number to a user.

        `Args:`
            user_id: string
                The id of the user.
            phone_type: string
                The type of the phone (e.g., "Home").
            phone: string
                The phone number.
        `Returns:`
            Phone json object
        """
        return self._base_post(
            endpoint="phone",
            exception_message="Could not create phone",
            user=f"/rest/v1/user/{user_id}/",
            phone_type=phone_type,
            phone=phone,
        )

    def delete_actionfield(self, actionfield_id):
        """
        Delete an actionfield.

        `Args:`
            actionfield_id: int
                The id of the actionfield to delete
        `Returns:`
            ``None``
        """

        resp = self.conn.delete(self._base_endpoint("actionfield", actionfield_id))
        logger.info(f"{resp.status_code}: {actionfield_id}")

    def update_user(self, user_id, **kwargs):
        """
        Update a user.

        `Args:`
            user_id: int
                The user id of the person to update
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns:`
            ``HTTP response from the patch request``
        """

        resp = self.conn.patch(self._base_endpoint("user", user_id), data=json.dumps(kwargs))
        logger.info(f"{resp.status_code}: {user_id}")

        return resp

    def update_phone(self, phone_id, **kwargs):
        """
        Update a phone record.

        `Args:`
            phone_id: int
                The phone id of the phone to update
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                at the /rest/v1/phone/schema/ path on any ActionKit instance.
        `Returns:`
            ``HTTP response from the patch request``
        """

        resp = self.conn.patch(self._base_endpoint("phone", phone_id), data=json.dumps(kwargs))
        logger.info(f"{resp.status_code}: {phone_id}")

        return resp

    def get_event(self, event_id):
        """Get an event.

        `Args:`
            event_id: int
                The id for the event.
        `Returns:`
            dict
                Event json object.

        """
        return self._base_get(f"event/{event_id}")

    def get_events(self, limit=None, **kwargs):
        """Get multiple events.

        `Args:`
            limit: int
                The number of events to return. If omitted, all events are returned.
            **kwargs:
                Optional arguments to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.

                Additionally, expressions to filter the data can also be provided. For addition
                info, visit `Django's docs on field lookups <https://docs.djangoproject.com/\
                en/3.1/topics/db/queries/#field-lookups>`_.

                .. code-block:: python

                    ak.get_events(name__contains="FirstName")
        `Returns:`
            Parsons.Table
                The events data.
        """
        return self.paginated_get("event", limit=limit, **kwargs)

    def update_event(self, event_id, **kwargs):
        """
        Update an event.

        `Args:`
            event_id: int
                The event id of the event to update
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns:`
            ``None``
        """

        resp = self.conn.patch(self._base_endpoint("event", event_id), data=json.dumps(kwargs))
        logger.info(f"{resp.status_code}: {event_id}")

    def create_event_field(self, event_id, name, value):
        """
        Create an event field (custom field on an event). Note that if an event
        field with this name already exists, this will add a second record.

        `Args:`
            event_id: int
                The id for the event.
            name: string
                The name of the event field.
            value: string
                The value of the event field.
        `Returns:`
            Event field json object
        """
        return self._base_post(
            endpoint="eventfield",
            exception_message="Could not create event field",
            event=f"/rest/v1/event/{event_id}/",
            name=name,
            value=value,
        )

    def update_event_field(self, eventfield_id, name, value):
        """
        Update an event field.

        `Args:`
            eventfield_id: int
                The id of the event field to update.
            name: string
                The name of the event field.
            value: string
                The value of the event field.
        `Returns:`
            ``None``
        """
        resp = self.conn.patch(
            self._base_endpoint("eventfield", eventfield_id),
            data=json.dumps(
                {
                    "name": name,
                    "value": value,
                }
            ),
        )
        logger.info(f"{resp.status_code}: {eventfield_id}")

    def get_blackholed_email(self, email):
        """
        Get a blackholed email. A blackholed email is an email that has been prevented from
        receiving bulk and transactional emails from ActionKit. `Documentation <https://\
        docs.actionkit.com/docs/manual/guide/mailings_tools.html#blackhole>`_.

        `Args:`
            email: str
                Blackholed email of the record to get.
        `Returns`:
            Parsons.Table
                The blackholed email data.
        """

        return self.paginated_get("blackholedemail", email=email)

    def blackhole_email(self, email):
        """
        Prevent an email from receiving bulk and transactional emails from ActionKit.
        `Documentation <https://docs.actionkit.com/docs/manual/guide/\
        mailings_tools.html#blackhole>`_.

        `Args:`
            user_id: str
                Email to blackhole
        `Returns:`
            API location of new resource
        """

        return self._base_post(
            endpoint="blackholedemail",
            exception_message="Could not blackhole email",
            email=email,
        )

    def delete_user_data(self, email, **kwargs):
        """
        Delete user data.

        `Args:`
            email: str
                Email of user to delete data
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://docs.actionkit.com/docs/manual/api/\
                rest/users.html>`_.
        `Returns:`
            API location of anonymized user
        """

        return self._base_post(
            endpoint="eraser",
            exception_message="Could not delete user data",
            email=email,
            **kwargs,
        )

    def delete_user(self, user_id):
        """
        Delete a user.

        `Args:`
            user_id: int
                The user id of the person to delete
        `Returns:`
            ``None``
        """

        resp = self.conn.delete(self._base_endpoint("user", user_id))
        logger.info(f"{resp.status_code}: {user_id}")

    def get_campaign(self, campaign_id):
        """
        Get a campaign.

        `Args:`
            campaign_id: int
                The campaign id of the record.
        `Returns`:
            Campaign json object
        """

        return self._base_get(
            endpoint="campaign",
            entity_id=campaign_id,
            exception_message="Campaign not found",
        )

    def get_campaign_fields(self):
        """
        Get list of valid campaign fields that can be passed with the
        :meth:`ActionKit.create_campaign` and :meth:`ActionKit.update_campaign` methods.

        `Returns`:
            List of campaign fields
        """

        resp = self._base_get(endpoint="campaign/schema")
        return list(resp["fields"].keys())

    def create_campaign(self, name, **kwargs):
        """
        Create a campaign.

        `Args:`
            name: str
                The name of the campaign to create
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns`:
            API location of new resource
        """

        return self._base_post(
            endpoint="campaign",
            exception_message="Could not create campaign",
            name=name,
            **kwargs,
        )

    def search_events_in_campaign(
        self,
        campaign_id,
        limit=None,
        order_by="id",
        ascdesc="asc",
        filters=None,
        exclude=None,
        **kwargs,
    ):
        """
        Get events in a campaign, with optional search filters.

        `Args:`
            campaign_id: int
                The id of the event campaign.
            limit: int
                The maximum number of objects to return.
            order_by: string
                Event attribute to order the results by. Defaults to id, which will normally
                be equivalent to ordering by created_at. See `ActionKit's docs on ordering
                <https://roboticdogs.actionkit.com/docs//manual/api/rest/overview.html#ordering>`_.
            ascdesc: string
                If "asc" (the default), returns events ordered by the attribute specified by
                the order_by parameter. If "desc", returns events in reverse order.
            filters: dictionary
                A dictionary for filtering by the attributes of the event or related object.
                Not all attributes are available for filtering, but an eventfield will work.
                For additional info, visit `Django's docs on field lookups
                <https://docs.djangoproject.com/en/3.1/topics/db/queries/#field-lookups>`_ and
                `ActionKit's docs on the search API
                <https://roboticdogs.actionkit.com/docs/manual/api/rest/examples/eventsearch.html>`_.

                .. code-block:: python

                    {
                        "title": "Example Event Title",
                        "field__name": "example_event_field_name",
                        "field__value": "Example event field value",
                    }
            exclude: dictionary
                A dictionary for excluding by the attributes of the event or related object.
                Uses the same format as the filters argument.
            **kwargs:
                A dictionary of other options for filtering. See `ActionKit's docs on the
                search API
                <https://roboticdogs.actionkit.com/docs/manual/api/rest/examples/eventsearch.html>`_.
        `Returns:`
            Parsons.Table
                The list of events.
        """
        if filters:
            for field, value in filters.items():
                kwargs[f"filter[{field}]"] = value
        if exclude:
            for field, value in exclude.items():
                kwargs[f"exclude[{field}]"] = value
        if ascdesc == "asc":
            kwargs["order_by"] = order_by
        else:
            kwargs["order_by"] = f"-{order_by}"
        return self.paginated_get(
            f"campaign/{campaign_id}/event_search",
            limit=limit,
            **kwargs,
        )

    def get_event_create_page(self, event_create_page_id):
        """
        Get a event create page.

        `Args:`
            event_create_page_id: int
                The event create page id of the record to get.
        `Returns`:
            Event create page json object
        """

        return self._base_get(
            endpoint="eventcreatepage",
            entity_id=event_create_page_id,
            exception_message="Event create page not found",
        )

    def get_event_create_page_fields(self):
        """
        Get list of event create page fields that can be passed with the
        :meth:`ActionKit.create_event_create_page`.

        `Returns`:
            List of event create page fields
        """

        resp = self._base_get(endpoint="eventcreatepage/schema")
        return list(resp["fields"].keys())

    def create_event_create_page(self, name, campaign_id, title, **kwargs):
        """
        Add an event page to a campaign.

        `Args:`
            campaign_id: int
                The campaign to assoicate page with
            name: str
                The name of the page to create
            title: str
                The title of the page to create
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns`:
            API location of new resource
        """

        return self._base_post(
            endpoint="eventcreatepage",
            exception_message="Could not create event create page",
            campaign=f"/rest/v1/campaign/{campaign_id}/",
            name=name,
            title=title,
            **kwargs,
        )

    def get_event_create_form(self, event_create_form_id):
        """
        Get a event create form.

        `Args:`
            event_create_form_id: int
                The event create form id of the record to get.
        `Returns`:
            Event create form json object
        """

        return self._base_get(
            endpoint="eventcreateform",
            entity_id=event_create_form_id,
            exception_message="Event create page not found",
        )

    def get_event_create_form_fields(self):
        """
        Get list of valid event create form fields that can be passed with the
        :meth:`ActionKit.create_event_create_form` method.

        `Returns`:
            List of event create form fields
        """

        resp = self._base_get(endpoint="eventcreateform/schema")
        return list(resp["fields"].keys())

    def create_event_create_form(self, page_id, thank_you_text, **kwargs):
        """
        Create a event create form.

        `Args:`
            page_id: int
                The page to associate the form with
            thank_you_text: str
                Free form thank you text
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns:`
            API location of new resource
        """

        return self._base_post(
            endpoint="eventcreateform",
            exception_message="Could not event create form",
            page=f"/rest/v1/eventcreatepage/{page_id}/",
            thank_you_text=thank_you_text,
            **kwargs,
        )

    def get_event_signup_page(self, event_signup_page_id):
        """
        Get event signup page.

        `Args:`
            event_signup_page_id: int
                The event signup page id of the record to get.
        `Returns`:
            Event signup page json object
        """

        return self._base_get(
            endpoint="eventsignuppage",
            entity_id=event_signup_page_id,
            exception_message="User page signup page not found",
        )

    def get_event_signup_page_fields(self):
        """
        Get list of valid event signup page fields that can be passed with the
        :meth:`ActionKit.create_event_signup_page` method.

        `Returns`:
            List of event signup page fields
        """

        resp = self._base_get(endpoint="eventsignuppage/schema")
        return list(resp["fields"].keys())

    def create_event_signup_page(self, name, campaign_id, title, **kwargs):
        """
        Add an event signup page to a campaign.

        `Args:`
            campaign_id: int
                The campaign to assoicate page with
            name: str
                The name of the page to create
            title: str
                The title of the page to create
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns`:
            API location of new resource
        """

        return self._base_post(
            endpoint="eventsignuppage",
            exception_message="Could not create signup page",
            campaign=f"/rest/v1/campaign/{campaign_id}/",
            name=name,
            title=title,
            **kwargs,
        )

    def get_event_signup_form(self, event_signup_form_id):
        """
        Get a user.

        `Args:`
            event_signup_form_id: str
                The event signup form id of the record to get.
        `Returns`:
            Event signup form json object
        """

        return self._base_get(
            endpoint="eventsignupform",
            entity_id=event_signup_form_id,
            exception_message="User page signup form not found",
        )

    def get_event_signup_form_fields(self):
        """
        Get list of valid event signup form fields that can be passed with the
        :meth:`ActionKit.create_event_signup_form` method.

        `Returns`:
            List of event signup form fields
        """

        resp = self._base_get(endpoint="eventsignupform/schema")
        return list(resp["fields"].keys())

    def create_event_signup_form(self, page_id, thank_you_text, **kwargs):
        """
        Create a event signup form.

        `Args:`
            page_id: int
                The page to associate the form with
            thank_you_text: str
                Free form thank you text
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns:`
            API location of new resource
        """

        return self._base_post(
            endpoint="eventsignupform",
            exception_message="Could not event create signup form",
            page=f"/rest/v1/page/{page_id}/",
            thank_you_text=thank_you_text,
            **kwargs,
        )

    def update_event_signup(self, event_signup_id, **kwargs):
        """
        Update an event signup.

        `Args:`
            event_signup_id: int
                The id of the event signup to update
            event_signup_dict: dict
                A dictionary of fields to update for the event signup.
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns:`
            ``None``
        """

        resp = self.conn.patch(
            self._base_endpoint("eventsignup", event_signup_id), data=json.dumps(kwargs)
        )
        logger.info(f"{resp.status_code}: {event_signup_id}")

    def get_mailer(self, entity_id):
        """
        Get a mailer.

        `Args:`
            entity_id: int
                The entity id of the record to get.
        `Returns`:
            Mailer json object
        """

        return self._base_get(endpoint="mailer", entity_id=entity_id)

    def create_mailer(self, **kwargs):
        """
        Create a mailer.

        `Args:`
            **kwargs:
                Arguments and fields to pass to the client. A full list can be found in the
                `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/manual/api/\
                rest/mailer.html>`_.
        `Returns:`
            URI of new mailer
        """

        return self._base_post(
            endpoint="mailer", exception_message="Could not create mailer", **kwargs
        )

    def copy_mailer(self, mailer_id):
        """
        copy a mailer
        returns new copy of mailer which should be updatable.
        """
        resp = self.conn.post(self._base_endpoint("mailer", entity_id=mailer_id) + "/copy")
        return resp

    def update_mailing(self, mailer_id, **kwargs):
        """
        Update a mailing.

        `Args:`
            mailing_id: int
                The id of the mailing to update
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns:`
            ``HTTP response from the patch request``
        """

        resp = self.conn.patch(self._base_endpoint("mailer", mailer_id), data=json.dumps(kwargs))
        logger.info(f"{resp.status_code}: {mailer_id}")
        return resp

    def rebuild_mailer(self, mailing_id):
        """
        Rebuild a mailer.

        `Args:`
            mailing_id: int
                Id of the mailer.
        `Returns:`
            URI to poll for progress
        """

        return self._base_post(
            endpoint="mailer/" + str(mailing_id) + "/rebuild",
            exception_message="Could not rebuild mailer",
        )

    def queue_mailer(self, mailing_id):
        """
        Queue a mailer.

        `Args:`
            mailing_id: int
                Id of the mailer.
        `Returns:`
            URI to poll for progress
        """

        return self._base_post(
            endpoint="mailer/" + str(mailing_id) + "/queue",
            exception_message="Could not queue mailer",
        )

    def paginated_get(self, object_type, limit=None, **kwargs):
        """Get multiple objects of a given type.

        `Args:`
            object_type: string
                The type of object to search for.
            limit: int
                The number of objects to return. If omitted, all objects are returned.
            **kwargs:
                Optional arguments to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.

                Additionally, expressions to filter the data can also be provided. For addition
                info, visit `Django's docs on field lookups <https://docs.djangoproject.com/\
                en/3.1/topics/db/queries/#field-lookups>`_.

                .. code-block:: python

                    ak.paginated_get(name__contains="FirstName")
        `Returns:`
            Parsons.Table
                The objects data.
        """
        # "The maximum number of objects returned per request is 100. Use paging
        # to get more objects."
        # (https://roboticdogs.actionkit.com/docs//manual/api/rest/overview.html#ordering)
        # get only `limit` objects if it's below 100, otherwise get 100 at a time
        kwargs["_limit"] = min(100, limit or 1_000_000_000)
        json_data = self._base_get(object_type, params=kwargs)
        data = json_data["objects"]

        next_url = json_data.get("meta", {}).get("next")
        while next_url:
            resp = self.conn.get(f"https://{self.domain}{next_url}")
            data.extend(resp.json().get("objects", []))
            next_url = resp.json().get("meta", {}).get("next")
            if limit and len(data) >= limit:
                break

        return Table(data[:limit])

    def paginated_get_custom_limit(
        self,
        object_type,
        limit=None,
        threshold_field=None,
        threshold_value=None,
        ascdesc="asc",
        **kwargs,
    ):
        """Get multiple objects of a given type, stopping based on the value of a field.

        `Args:`
            object_type: string
                The type of object to search for.
            limit: int
                The maximum number of objects to return. Even if the threshold
                value is not reached, if the limit is set, then at most this many
                objects will be returned.
            threshold_field: string
                The field used to determine when to stop.
                Must be one of the options for ordering by.
            threshold_value: string
                The value of the field to stop at.
            ascdesc: string
                If "asc" (the default), return all objects below the threshold value.
                If "desc", return all objects above the threshold value.
            **kwargs:
                You can also add expressions to filter the data beyond the limit/threshold values
                above. For additional info, visit `Django's docs on field lookups
                <https://docs.djangoproject.com/en/3.1/topics/db/queries/#field-lookups>`_.

                .. code-block:: python

                    ak.paginated_get(name__contains="FirstName")
        `Returns:`
            Parsons.Table
                The objects data.
        """
        # "The maximum number of objects returned per request is 100. Use paging
        # to get more objects."
        # (https://roboticdogs.actionkit.com/docs//manual/api/rest/overview.html#ordering)
        kwargs["_limit"] = min(100, limit or 1_000_000_000)
        if ascdesc == "asc":
            kwargs["order_by"] = threshold_field
        else:
            kwargs["order_by"] = "-" + threshold_field
        json_data = self._base_get(object_type, params=kwargs)
        data = json_data["objects"]
        next_url = json_data.get("meta", {}).get("next")
        while next_url:
            last = data[-1].get(threshold_field)
            if ascdesc == "asc" and last > threshold_value:
                break
            if ascdesc == "desc" and last < threshold_value:
                break
            resp = self.conn.get(f"https://{self.domain}{next_url}")
            data += resp.json().get("objects", [])
            next_url = resp.json().get("meta", {}).get("next")
            if limit and len(data) >= limit:
                break
        # This could be more efficient but it's still O(n) so no big deal
        i = len(data) - 1  # start at the end; 0-indexed means the end is length - 1
        if ascdesc == "asc":
            while data[i].get(threshold_field) > threshold_value:
                i = i - 1
        else:
            while data[i].get(threshold_field) < threshold_value:
                i = i - 1
        data = data[:i]
        return Table(data[:limit])

    def get_order(self, order_id):
        """
        Get an order.

        `Args:`
            order_id: int
                The order id of the record to get.
        `Returns`:
            User json object
        """

        return self._base_get(
            endpoint="order", entity_id=order_id, exception_message="Order not found"
        )

    def update_order(self, order_id, **kwargs):
        """
        Update an order.

        `Args:`
            order_id: int
                The id of the order to update
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns:`
            ``None``
        """

        resp = self.conn.patch(self._base_endpoint("order", order_id), data=json.dumps(kwargs))
        logger.info(f"{resp.status_code}: {order_id}")

    def update_order_user_detail(self, user_detail_id, **kwargs):
        """
        Update an order user detail.

        `Args:`
            user_detail_id: int
                The id of the order user detail to update
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                at the /rest/v1/orderuserdetail/schema/ path on any ActionKit instance.
        `Returns:`
            ``HTTP response from the patch request``
        """

        resp = self.conn.patch(
            self._base_endpoint("orderuserdetail", user_detail_id), data=json.dumps(kwargs)
        )
        logger.info(f"{resp.status_code}: {user_detail_id}")

        return resp

    def get_orderrecurring(self, orderrecurring_id):
        """
        Get an orderrecurring.

        `Args:`
            orderrecurring_id: int
                The orderrecurring id of the record to get.
        `Returns`:
            User json object
        """

        return self._base_get(
            endpoint="orderrecurring",
            entity_id=orderrecurring_id,
            exception_message="Orderrecurring not found",
        )

    def cancel_orderrecurring(self, recurring_id):
        """
        Cancel a recurring order.

        `Args:`
            recurring_id: int
                The id of the recurring order to update (NOT the order_id)
        `Returns:`
            ``None``
        """

        resp = self.conn.post(self._base_endpoint("orderrecurring", str(recurring_id) + "/cancel"))
        logger.info(f"{resp.status_code}: {recurring_id}")
        return resp

    def update_orderrecurring(self, orderrecurring_id, **kwargs):
        """
        Update a recurring order.

        `Args:`
            orderrecurring_id: int
                The id of the orderrecurring to update
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns:`
            ``None``
        """

        resp = self.conn.patch(
            self._base_endpoint("orderrecurring", orderrecurring_id),
            data=json.dumps(kwargs),
        )
        logger.info(f"{resp.status_code}: {orderrecurring_id}")

    def get_orders(self, limit=None, **kwargs):
        """Get multiple orders.

        `Args:`
            limit: int
                The number of orders to return. If omitted, all orders are returned.
            **kwargs:
                Optional arguments to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.

                Additionally, expressions to filter the data can also be provided. For addition
                info, visit `Django's docs on field lookups <https://docs.djangoproject.com/\
                en/3.1/topics/db/queries/#field-lookups>`_.

                .. code-block:: python

                    ak.get_orders(import_id="my-import-123")
        `Returns:`
            Parsons.Table
                The orders data.
        """
        return self.paginated_get("order", limit=limit, **kwargs)

    def update_paymenttoken(self, paymenttoken_id, **kwargs):
        """
        Update a saved payment token.

        `Args:`
            paymenttoken_id: int
                The id of the payment token to update
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns:`
            ``HTTP response``
        """

        resp = self.conn.patch(
            self._base_endpoint("paymenttoken", paymenttoken_id),
            data=json.dumps(kwargs),
        )
        logger.info(f"{resp.status_code}: {paymenttoken_id}")
        return resp

    def get_page_followup(self, page_followup_id):
        """
        Get a page followup.

        `Args:`
            page_followup_id: int
                The user id of the record to get.
        `Returns`:
            Page followup json object
        """

        return self._base_get(
            endpoint="pagefollowup",
            entity_id=page_followup_id,
            exception_message="Page followup not found",
        )

    def get_page_followup_fields(self):
        """
        Get list of valid page followup fields that can be passed with the
        :meth:`ActionKit.create_page_followup` method.

        `Returns`:
            List of page followup fields
        """

        resp = self._base_get(endpoint="pagefollowup/schema")
        return list(resp["fields"].keys())

    def create_page_followup(self, signup_page_id, url, **kwargs):
        """
        Add a page followup.

        `Args:`
            signup_page_id: int
                The signup page to associate the followup page with
            url: str
                URL of the folloup page
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns`:
            API location of new resource
        """

        return self._base_post(
            endpoint="pagefollowup",
            exception_message="Could not create page followup",
            page=f"/rest/v1/eventsignuppage/{signup_page_id}/",
            url=url,
            **kwargs,
        )

    def get_survey_question(self, survey_question_id):
        """
        Get a survey question.

        `Args:`
            survey_question_id: int
                The survey question id of the record to get.
        `Returns`:
            Survey question json object
        """

        return self._base_get(
            endpoint="surveyquestion",
            entity_id=survey_question_id,
            exception_message="Survey question not found",
        )

    def update_survey_question(self, survey_question_id, **kwargs):
        """
        Update a survey question.

        `Args:`
            survey_question_id: int
                The id of the survey question to update
            survey_question_dict: dict
                A dictionary of fields to update for the survey question.
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns:`
            ``None``
        """

        resp = self.conn.patch(
            self._base_endpoint("surveyquestion", survey_question_id),
            data=json.dumps(kwargs),
        )
        logger.info(f"{resp.status_code}: {survey_question_id}")

    def create_transaction(self, **kwargs):
        """
        Create a transaction.

        `Args:`
            **kwargs:
                Optional arguments and fields to pass to the client.
        `Returns:`
            Transaction json object
        """

        return self._base_post(
            endpoint="transaction",
            exception_message="Could not create transaction",
            **kwargs,
        )

    def update_transaction(self, transaction_id, **kwargs):
        """
        Update a transaction.

        `Args:`
            transaction_id: int
                The id of the transaction to update
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns:`
            ``None``
        """

        resp = self.conn.patch(
            self._base_endpoint("transaction", transaction_id), data=json.dumps(kwargs)
        )
        logger.info(f"{resp.status_code}: {transaction_id}")

    def get_transactions(self, limit=None, **kwargs):
        """Get multiple transactions.

        `Args:`
            limit: int
                The number of transactions to return. If omitted, all transactions are returned.
            **kwargs:
                Optional arguments to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.

                Additionally, expressions to filter the data can also be provided. For addition
                info, visit `Django's docs on field lookups <https://docs.djangoproject.com/\
                en/3.1/topics/db/queries/#field-lookups>`_.

                .. code-block:: python

                    ak.get_transactions(order="order-1")
        `Returns:`
            Parsons.Table
                The transactions data.
        """
        return self.paginated_get("transaction", limit=limit, **kwargs)

    def create_generic_action(self, page, email=None, ak_id=None, **kwargs):
        """
        Post a generic action. One of ``ak_id`` or ``email`` is a required argument.

        `Args:`
            page:
                The page to post the action. The page short name.
            email:
                The email address of the user to post the action.
            ak_id:
                The action kit id of the record.
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                in the `ActionKit API Documentation <https://roboticdogs.actionkit.com/docs/\
                manual/api/rest/actionprocessing.html>`_.
        `Returns`:
            dict
                The response json
        """

        if not email or ak_id:
            raise ValueError("One of email or ak_id is required.")

        return self._base_post(
            endpoint="action",
            exception_message="Could not create action.",
            email=email,
            page=page,
            return_full_json=True,
            **kwargs,
        )

    def update_import_action(self, action_id, **kwargs):
        """
        Update an import action.

        `Args:`
            action_id: int
                The action id of the import action to update
            **kwargs:
                Optional arguments and fields to pass to the client. A full list can be found
                at the /rest/v1/importaction/schema/ path on any ActionKit instance.
        `Returns:`
            ``HTTP response from the patch request``
        """

        resp = self.conn.patch(
            self._base_endpoint("importaction", action_id), data=json.dumps(kwargs)
        )
        logger.info(f"{resp.status_code}: {action_id}")

        return resp

    def bulk_upload_csv(
        self,
        csv_file,
        import_page,
        autocreate_user_fields=False,
        user_fields_only=False,
    ):
        """
        Bulk upload a csv file of new users or user updates.
        If you are uploading a table object, use bulk_upload_table instead.
        See `ActionKit User Upload Documentation <https://roboticdogs.actionkit.com/docs/manual/api/rest/uploads.html>`_
        Be careful that blank values in columns will overwrite existing data.

        If you get a 500 error, try sending a much smaller file (say, one row),
        which is more likely to return the proper 400 with a useful error message

        `Args:`
            import_page: str
                The page to post the action. The page short name.
            csv_file: str or buffer
                The csv (optionally zip'd) file path or a file buffer object
                A user_id or email column is required.
                ActionKit rejects files that are larger than 128M
            autocreate_user_fields: bool
              When True columns starting with "user_" will be uploaded as user fields.
              See the `autocreate_user_fields documentation
              <https://roboticdogs.actionkit.com/docs/manual/api/rest/uploads.html#create-a-multipart-post-request>`_.
            user_fields_only: bool
              When uploading only an email/user_id column and user_ user fields,
              ActionKit has a fast processing path.
              This doesn't work, if you upload a zipped csv though.
        `Returns`:
            dict
                success: whether upload was successful
                progress_url: an API URL to get progress on upload processing
                res: requests http response object
        """

        # self.conn defaults to JSON, but this has to be form/multi-part....
        upload_client = self._conn({"accepts": "application/json"})
        if isinstance(csv_file, str):
            csv_file = open(csv_file, "rb")

        url = self._base_endpoint("upload")
        files = {"upload": csv_file}
        data = {
            "page": import_page,
            "autocreate_user_fields": int(autocreate_user_fields),
            "user_fields_only": int(user_fields_only),
        }
        with upload_client.post(url, files=files, data=data) as res:
            progress_url = res.headers.get("Location")
            rv = {
                "res": res,
                "success": res.status_code == 201,
                "id": progress_url.split("/")[-2] if progress_url else None,
                "progress_url": progress_url,
            }
            return rv

    def bulk_upload_table(
        self,
        table,
        import_page,
        autocreate_user_fields=0,
        no_overwrite_on_empty=False,
        set_only_columns=None,
    ):
        """
        Bulk upload a table of new users or user updates.
        See `ActionKit User Upload Documentation <https://roboticdogs.actionkit.com/docs/manual/api/rest/uploads.html>`_
        Be careful that blank values in columns will overwrite existing data.

        Tables with only an identifying column (user_id/email) and user_ user fields
        will be fast-processed -- this is useful for setting/updating user fields.

        .. note::
            If you get a 500 error, try sending a much smaller file (say, one row),
            which is more likely to return the proper 400 with a useful error message

        `Args:`
            import_page: str
                The page to post the action. The page short name.
            table: Table Class
                A Table of user data to bulk upload
                A user_id or email column is required.
            autocreate_user_fields: bool
                When True columns starting with "user_" will be uploaded as user fields.
                `ActionKit <https://actionkit.com/>`_.
                See the autocreate_user_fields `documentation <https://roboticdogs.actionkit.com/docs/manual/api/rest/uploads.html#create-a-multipart-post-request>`_.
            no_overwrite_on_empty: bool
                When uploading user data, ActionKit will, by default, take a blank value
                and overwrite existing data for that user.
                This can be undesirable, if the goal is to only send updates.
                Setting this to True will divide up the table into multiple upload
                batches, changing the columns uploaded based on permutations of
                empty columns.
            set_only_columns: list
                This is similar to no_overwrite_on_empty but restricts to a specific set of columns
                which, if blank, should not be overwritten.
        `Returns`:
            dict
                success: bool -- whether upload was successful (individual rows may not have been)
                results: [dict] -- This is a list of the full results.
                         progress_url and res for any results
        """

        import_page = check_env.check("ACTION_KIT_IMPORTPAGE", import_page)
        upload_tables = self._split_tables_no_empties(
            table, no_overwrite_on_empty, set_only_columns
        )
        results = []
        for tbl in upload_tables:
            user_fields_only = int(
                not any(h for h in tbl.columns if h != "email" and not h.startswith("user_"))
            )
            results.append(
                self.bulk_upload_csv(
                    tbl.to_csv(),
                    import_page,
                    autocreate_user_fields=autocreate_user_fields,
                    user_fields_only=user_fields_only,
                )
            )
        return {"success": all(r["success"] for r in results), "results": results}

    def _split_tables_no_empties(self, table, no_overwrite_on_empty, set_only_columns):
        table_groups = {}
        # uploading combo of user_id and email column should be mutually exclusive
        blank_columns_test = table.columns
        if not no_overwrite_on_empty:
            blank_columns_test = set(["user_id", "email"] + (set_only_columns or [])).intersection(
                table.columns
            )
        for row in table:
            blanks = tuple(k for k in blank_columns_test if row.get(k) in (None, ""))
            grp = table_groups.setdefault(blanks, [])
            grp.append(row)
        results = []
        for blanks, subset in table_groups.items():
            subset_table = Table(subset)
            if blanks:
                subset_table.table = subset_table.table.cutout(*blanks)
            logger.debug(f"Column Upload Blanks: {blanks}")
            logger.debug(f"Column Upload Columns: {subset_table.columns}")
            if not {"user_id", "email"}.intersection(subset_table.columns):
                logger.warning(
                    f"Upload will fail without user_id or email. "
                    f"Rows: {subset_table.num_rows}, Columns: {subset_table.columns}"
                )
            results.append(subset_table)
        return results

    def collect_upload_errors(self, result_array):
        """
        Collect any upload errors as a list of objects from bulk_upload_table 'results' key value.
        This waits for uploads to complete, so it may take some time if you uploaded a large file.
        `Args:`
            result_array: list
                After receiving a dict back from bulk_upload_table you may want to see if there
                were any errors in the uploads.  If you call collect_upload_errors(result_array)
                it will iterate across each of the uploads fetching the final result of e.g.
                /rest/v1/uploaderror?upload=123
        `Returns`:
            [dict]
                message: str -- error message
                upload: str -- upload progress API path e.g. "/rest/v1/upload/123456/"
                id: int -- upload error record id (different than upload id)
        """
        errors = []
        for res in result_array:
            upload_id = res.get("id")
            if upload_id:
                # Pend until upload is complete
                while True:
                    upload = self._base_get(endpoint="upload", entity_id=upload_id)
                    if upload.get("is_completed"):
                        break
                    else:
                        time.sleep(1)

                # ActionKit limits length of error list returned
                # Iterate until all errors are gathered
                error_count = upload.get("has_errors")
                limit = 20

                error_pages = math.ceil(error_count / limit)
                for page in range(0, error_pages):
                    error_data = self._base_get(
                        endpoint="uploaderror",
                        params={
                            "upload": upload_id,
                            "_limit": limit,
                            "_offset": page * limit,
                        },
                    )
                    logger.debug(f"error collect result: {error_data}")
                    errors.extend(error_data.get("objects", []))

        return errors



import re


def format_phone_number(phone_number, country_code="1"):
    """
    Formats a phone number in E.164 format, which is the international standard for phone numbers.
    Example: Converts "555-555-5555" -> "+15555555555"

    `Args:`
        phone_number (str):
            The phone number to be formatted.
        country_code (str):
            The country code to be used as a prefix.
        Defaults to "1" (United States).

    `Returns:`
        str:
            The formatted phone number in E.164 format.
    """
    # Remove non-numeric characters and leading zeros
    digits = re.sub(r"[^\d]", "", phone_number.lstrip("0"))

    # Check if the phone number is valid
    if len(digits) < 10:
        return None

    # Handle country code prefix
    if not digits.startswith(country_code):
        digits = country_code + digits

    # Format the phone number in E.164 format
    formatted_number = "+" + digits

    return formatted_number



import logging

from pyairtable import Api as client

from parsons.etl import Table
from parsons.utilities import check_env

logger = logging.getLogger(__name__)


class Airtable(object):
    """
    `Args:`
        base_key: str
            The key/ID of the Airtable base that you will interact with, typically
            prefixed with `app`.
        table_name: str
            The name or key/ID of the table in the base. The table name is the
            equivalent of the sheet name in Excel or GoogleDocs. The ID can be
            found in the URL and is typically prefixed with `tbl`.
        personal_access_token: str
            The Airtable personal access token. Not required if ``AIRTABLE_PERSONAL_ACCESS_TOKEN``
            env variable set.
    """

    def __init__(self, base_key, table_name, personal_access_token=None):
        self.personal_access_token = check_env.check(
            "AIRTABLE_PERSONAL_ACCESS_TOKEN", personal_access_token
        )
        self.client = client(self.personal_access_token).table(base_key, table_name)

    def get_record(self, record_id):
        """
        Returns a single record.

        `Args:`
            record_id: str
                The Airtable record `id`
        `Returns:`
            A dictionary of the record
        """

        return self.client.get(record_id)

    def get_records(
        self,
        fields=None,
        max_records=None,
        view=None,
        formula=None,
        sort=None,
        sample_size=None,
    ):
        """
        `Args:`
            fields: str or lst
                Only return specified column or list of columns. The column name is
                case sensitive
            max_records: int
                The maximum total number of records that will be returned.
            view: str
                If set, only the records in that view will be returned. The records will be
                sorted according to the order of the view.
            formula: str
                The formula will be evaluated for each record, and if the result
                is not 0, false, "", NaN, [], or #Error! the record will be included
                in the response.

                If combined with view, only records in that view which satisfy the
                formula will be returned. For example, to only include records where
                ``COLUMN_A`` isn't empty, pass in: ``"NOT({COLUMN_A}='')"``

                For more information see
                `Airtable Docs on formulas. <https://airtable.com/api>`_

                Usage - Text Column is not empty:

                ``airtable.get_all(formula="NOT({COLUMN_A}='')")``

                Usage - Text Column contains:

                ``airtable.get_all(formula="FIND('SomeSubText', {COLUMN_STR})=1")``

            sort: str or lst
                Specifies how the records will be ordered. If you set the view parameter, the
                returned records in that view will be sorted by these fields. If sorting by
                multiple columns, column names can be passed as a list. Sorting Direction is
                ascending by default, but can be reversed by prefixing the column name with a minus
                sign -.

                Example usage:
                ``airtable.get_records(sort=['ColumnA', '-ColumnB'])``

            sample_size: int
                Number of rows to sample before determining columns

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        if isinstance(fields, str):
            fields = [fields]
        # Raises an error if sort is None type. Thus, only adding if populated.
        kwargs = {
            "fields": fields,
            "max_records": max_records,
            "view": view,
            "formula": formula,
        }
        if sort:
            kwargs["sort"] = sort

        tbl = Table(self.client.all(**kwargs))

        # If the results are empty, then return an empty table.
        if "fields" not in tbl.columns:
            return Table([[]])

        unpack_dicts_kwargs = {
            "column": "fields",
            "prepend": False,
        }
        if fields:
            unpack_dicts_kwargs["keys"] = fields

        if sample_size:
            unpack_dicts_kwargs["sample_size"] = sample_size

        return tbl.unpack_dict(**unpack_dicts_kwargs)

    def insert_record(self, row, typecast=False):
        """
        Insert a single record into an Airtable.

        `Args:`
            row: dict
                Fields to insert. Must be dictionary with Column names as Key.
            typecast: boolean
                Automatic data conversion from string values.
        `Returns:`
            Dictionary of inserted row
        """

        resp = self.client.create(row, typecast=typecast)
        logger.info("Record inserted")
        return resp

    def insert_records(self, table, typecast=False):
        """
        Insert multiple records into an Airtable. The columns in your Parsons table must
        exist in the Airtable. The method will attempt to map based on column name, so the
        order of the columns is irrelevant.

        `Args:`
            table: A Parsons Table or list of dicts
                Insert a Parsons table or list
            typecast: boolean
                Automatic data conversion from string values.
        `Returns:`
            List of dictionaries of inserted rows
        """

        if isinstance(table, Table):
            table = table.to_dicts()

        resp = self.client.batch_create(table, typecast=typecast)
        logger.info(f"{len(table)} records inserted.")
        return resp

    def update_record(self, record_id, fields, typecast=False, replace=False):
        """
        Updates a record by its record `id`. Only Fields passed are updated, the rest are left as
        is.

        `Args:`
            record_id: str
                The Airtable record `id`
            fields: dict
                Fields to insert. Must be dictionary with Column names as Key.
            typecast: boolean
                Automatic data conversion from string values.
            replace: boolean
                Only provided fields are updated. If `True`, record is replaced in its
                entirety by provided fields; if a field is not included its value
                will bet set to null.
        `Returns:`
            Dictionary of updated row
        """

        resp = self.client.update(record_id, fields, typecast=typecast, replace=replace)
        logger.info(f"{record_id} updated")
        return resp

    def update_records(self, table, typecast=False, replace=False):
        """
        Update multiple records into an Airtable. The columns in your Parsons table must
        exist in the Airtable, and the record `id` column must be present. The method
        will attempt to map based on column name, so the order of the columns is irrelevant.

        `Args:`
            table: A Parsons Table or list of dicts
                Insert a Parsons table or list. Record must contain the record `id` column
                and columns containing the fields to update
            typecast: boolean
                Automatic data conversion from string values.
            replace: boolean
                Only provided fields are updated. If `True`, record is replaced in its
                entirety by provided fields; if a field is not included its value
                will bet set to null.
        `Returns:`
            List of dicts of updated records
        """

        # the update/upsert API call expects a dict/object shape of:
        # { id: string, fields: { column_name: value, ... } }
        # the map_update_fields helper will convert the flat table field
        # columns/keys into this nested structure
        table = list(map(map_update_fields, table))

        resp = self.client.batch_update(table, typecast=typecast, replace=replace)
        logger.info(f"{len(resp)} records updated.")
        return resp

    def upsert_records(self, table, key_fields=None, typecast=False, replace=False):
        """
        Update and/or create records, either using `id` (if included) or using a set of
        fields (`key_fields`) to look for matches. The columns in your Parsons table must
        exist in the Airtable. The method will attempt to map based on column name,
        so the order of the columns is irrelevant.

        `Args:`
            table: A Parsons Table or list of dicts
                Parsons table or list with records to upsert. Records must contain the record
                `id` column or the column(s) defined in `key_fields`.
            key_fields: list of str
              List of field names that Airtable should use to match records in the input
              with existing records.
            typecast: boolean
                Automatic data conversion from string values.
            replace: boolean
                Only provided fields are updated. If `True`, record is replaced in its
                entirety by provided fields; if a field is not included its value
                will bet set to null.
        `Returns:`
            Dictionary containing:
                - `updated_records`: list of updated record `id`s
                - `created_records`: list of created records `id`s
                - `records`: list of records
        """

        # the update/upsert API call expects a dict/object shape of:
        # { id: string, fields: { column_name: value, ... } }
        # the map_update_fields helper will convert the flat table field
        # columns/keys into this nested structure
        table = list(map(map_update_fields, table))

        resp = self.client.batch_upsert(table, key_fields, typecast=typecast, replace=replace)

        updated_records = resp["updatedRecords"]
        created_records = resp["createdRecords"]

        logger.info(
            f"{len(updated_records)} records updated, {len(created_records)} records created."
        )

        return {
            "records": resp["records"],
            "updated_records": updated_records,
            "created_records": created_records,
        }

    def delete_record(self, record_id):
        """
        Deletes a record by its record `id`.

        `Args:`
            record_id: str
                The Airtable record `id`
        `Returns:`
            Dictionary of record `id` and `deleted` status
        """

        resp = self.client.delete(record_id)
        logger.info(f"{record_id} updated")
        return resp

    def delete_records(self, table):
        """
        Delete multiple records from an Airtable.

        `Args:`
            table: A Parsons Table or list containing the record `id`s to delete.
        `Returns:`
            List of dicts with record `id` and `deleted` status
        """

        if isinstance(table, Table):
            table = table.to_dicts()

        # the API expects a list of ids which this method can accept directly;
        # otherwise if a table or list of dicts containing the `id` key/column
        # is provided then map the ids into the expected list of id strings.

        if any(isinstance(row, dict) for row in table):
            table = [row["id"] for row in table]

        resp = self.client.batch_delete(table)
        logger.info(f"{len(table)} records deleted.")
        return resp


def map_update_fields(record):
    record_id = record.get("id")
    if "id" in record:
        del record["id"]
        return {"id": record_id, "fields": record}

    return {"fields": record}



import logging

import parsons.databases.database.constants as consts

logger = logging.getLogger(__name__)


class DatabaseCreateStatement:
    def __init__(self):
        self.INT_TYPES = consts.INT_TYPES
        self.SMALLINT = consts.SMALLINT
        self.SMALLINT_MIN = consts.SMALLINT_MIN
        self.SMALLINT_MAX = consts.SMALLINT_MAX
        self.MEDIUMINT = consts.MEDIUMINT
        self.MEDIUMINT_MIN = consts.MEDIUMINT_MIN
        self.MEDIUMINT_MAX = consts.MEDIUMINT_MAX
        self.INT = consts.INT
        self.INT_MIN = consts.INT_MIN
        self.INT_MAX = consts.INT_MAX
        self.BIGINT = consts.BIGINT
        self.FLOAT = consts.FLOAT

        self.BOOL = consts.BOOL

        self.VARCHAR = consts.VARCHAR
        self.RESERVED_WORDS = consts.RESERVED_WORDS
        self.COL_NAME_MAX_LEN = consts.COL_NAME_MAX_LEN
        self.IS_CASE_SENSITIVE = consts.IS_CASE_SENSITIVE
        self.REPLACE_CHARS = consts.REPLACE_CHARS

    # This will allow child classes to modify how these columns are handled.
    def _rename_reserved_word(self, col, index=None):
        """Return the renamed column.

        `Args`:
            col: str
                The column to rename.
            index: int
                (Optional) The index of the column.
        `Returns`:
            str
                The rename column.
        """
        return f"{col}_"

    def _rename_duped(self, col, index):
        """Return the renamed column.

        `Args`:
            col: str
                The column to rename.
            index: int
                (Optional) The index of the column.
        `Returns`:
            str
                The rename column.
        """
        return f"{col}_{index}"

    def get_bigger_int(self, int1, int2):
        """Return the bigger of the two ints.

        `Args`:
            int1: str
                The string representation if an int type.
            int2: str
                The string representation if an int type.
        `Returns`:
            str
                A string representation of the higher of the two int types.
        """
        WEIGHTS = {
            self.SMALLINT: 100,
            self.MEDIUMINT: 200,
            self.INT: 300,
            self.BIGINT: 400,
        }

        return int1 if WEIGHTS.get(int1, -1) >= WEIGHTS.get(int2, -1) else int2

    def is_valid_sql_num(self, val):
        """Check whether val is a valid sql number.

        `Args`:
            val: any
                The values to check.
        `Returns`:
            bool
                Whether or not the value is a valid sql number.
        """
        # Python accepts numbers with single-underscore separators such as
        # 100_000 (evals to 100000)
        # SQL engines may not support this

        # If its python type is a number, then it's a number
        # 100_000 is stored at 100000 in memory so it's fine
        # not using `isinstance` b/c isinstance(bool. (int, float)) == True
        if type(val) in (int, float):
            return True

        # If it can be cast to a number and it doesn't contain underscores
        # then it's a valid sql number
        # Also check the first character is not zero
        try:
            if (float(val) or 1) and "_" not in val and (val in ("0", "0.0") or val[0] != "0"):
                return True
            else:
                return False

        # If it can't be cast to a number in python
        # then it's not a number in sql
        except (TypeError, ValueError):
            return False

    def detect_data_type(self, value, cmp_type=None):
        """Detect the higher of value's type cmp_type.

        1. check if it's a string
        2. check if it's a number
          a. check if it's a float
          b. check if it's an int

        `Args`:
            value: str
                The value to inspect.
            cmp_type: str
                The string representation of a type to compare with
                ``value``'s type.
        `Returns`:
            str
                The string representation of the higher of the two types.
        """
        # Stop if the compare type is already a varchar
        # varchar is the highest data type.
        if cmp_type == self.VARCHAR:
            result = cmp_type

        elif isinstance(value, bool):
            result = self.BOOL

        elif value is None:
            result = cmp_type

        # Make sure that it is a valid integer
        # Python accepts 100_000 as a valid form of 100000,
        # however a sql engine may throw an error
        elif not self.is_valid_sql_num(value):
            result = self.VARCHAR

        elif isinstance(value, float) or cmp_type == self.FLOAT:
            result = self.FLOAT

        # The value is very likely an int
        # let's get its size
        # If the compare types are empty and use the types of the current value
        elif isinstance(value, int) and cmp_type in (self.INT_TYPES + [None, "", self.BOOL]):
            # Use smallest possible int type above TINYINT
            if self.SMALLINT_MIN < value < self.SMALLINT_MAX:
                result = self.get_bigger_int(self.SMALLINT, cmp_type)
            elif self.MEDIUMINT_MIN < value < self.MEDIUMINT_MAX:
                result = self.get_bigger_int(self.MEDIUMINT, cmp_type)
            elif self.INT_MIN < value < self.INT_MAX:
                result = self.get_bigger_int(self.INT, cmp_type)
            else:
                result = self.BIGINT

        else:
            # Need to determine who makes it all the way down here
            logger.debug(f"Unexpected object type: {type(value)}")
            result = cmp_type

        return result

    def format_column(self, col, index="", replace_chars=None, col_prefix="_"):
        """Format the column to meet database contraints.

        Formats the columns as follows:
            1. Coverts to lowercase (if case insensitive)
            2. Strips leading and trailing whitespace
            3. Replaces invalid characters
            4. Renames if in reserved words
        `Args`:
            col: str
                The column to format.
            index: int
                (Optional) The index of the column. Used if the column is empty.
            replace_chars: dict
                A dictionary of invalid characters and their replacements. If
                ``None`` uses {" ": "_"}
            col_prefix: str
                The prefix to use when the column is empty or starts with an
                invalid character.
        `Returns`:
            str
                The formatted column.
        """
        replace_chars = replace_chars or self.REPLACE_CHARS

        if not self.IS_CASE_SENSITIVE:
            col = col.lower()

        col = col.strip()

        for char, rep_char in replace_chars.items():
            col = col.replace(char, rep_char)

        # The format for this column passes all other checks
        if col == "":
            return f"{col_prefix}{index}"

        if col.upper() in self.RESERVED_WORDS:
            col = self._rename_reserved_word(col, index)

        if col[0].isdigit():
            col = f"x_{col}"

        if len(col) > self.COL_NAME_MAX_LEN:
            col = col[: self.COL_NAME_MAX_LEN]

        return col

    def format_columns(self, cols, **kwargs):
        """Format the columns to meet database contraints.

        This method relies on ``format_column`` to handle most changes. It
        only handles duplicated columns. Options to ``format_column`` can be
        passed through kwargs.

        `Args`:
            cols: list
                The columns to format.
            kwargs: dicts
                Keyword arguments to pass to ``format_column``.
        `Returns`:
            list
                The formatted columns.
        """
        formatted_cols = []

        for idx, col in enumerate(cols):
            formatted_col = self.format_column(col, index=idx, **kwargs)

            if formatted_col in formatted_cols:
                formatted_col = self._rename_duped(formatted_col, idx)

            formatted_cols.append(formatted_col)

        return formatted_cols
